import subprocess
import streamlit as st
import ast
from io import StringIO
import yaml
from ovo import Design, Pool, Project, DesignJob, DescriptorJob, Round
from ovo import db, get_username, config

st.set_page_config(layout="wide", page_title="Debug", page_icon="🖥️")

st.title("🧬 Debug")

st.write("This page is visible only to selected admin users.")

st.subheader("Run shell command")
with st.form(key="run_command", border=False):
    shell_command = st.text_area("Enter command:", placeholder="ls -la", key="shell_command", height=80)
    if st.form_submit_button("Run"):
        with st.spinner("Running command..."):
            result = subprocess.run(
                shell_command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                shell=True,
            )
        st.write("Finished with output:" if result.returncode == 0 else "Finished with error:")
        st.code(result.stdout)

st.subheader("Run python command")
with st.form(key="run_python", border=False):
    python_command = st.text_area(
        "Enter command:", key="python_command", placeholder='pool = db.get(Pool, "abc");\nprint(pool.name)', height=150
    )
    if st.form_submit_button("Run"):
        with st.spinner("Running command..."):
            result = subprocess.run(
                f"""
                python <<EOF
from ovo import *
{python_command}
EOF
                """,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                shell=True,
            )
            st.write("Command:")
            st.code(python_command, language="python")
        st.write("Finished with output:" if result.returncode == 0 else "Finished with error:")
        st.code(result.stdout)

st.subheader("Session state")

st.write(st.session_state)

st.subheader("Config")

s = StringIO()
yaml.dump(config.model_dump(), s, sort_keys=False)
s.seek(0)
st.code(s.getvalue())

st.subheader("Database")

st.write("#### Projects")
kwargs = st.text_input("Filter JSON", placeholder='{"name": "Personal project"}', key="project_kwargs")
try:
    kwargs = ast.literal_eval(kwargs)
except:
    if kwargs:
        st.warning("Invalid kwargs format. Please enter a valid dictionary.")
    kwargs = {}
project_df = db.select_dataframe(model=Project, limit=1000, **kwargs)
st.dataframe(project_df)

st.write("#### Rounds")
kwargs = st.text_input("Filter JSON", key="rounds_kwargs")
try:
    kwargs = ast.literal_eval(kwargs)
except:
    if kwargs:
        st.warning("Invalid kwargs format. Please enter a valid dictionary.")
    kwargs = {}
rounds_df = db.select_dataframe(model=Round, limit=1000, **kwargs)
st.dataframe(rounds_df)

st.write("#### Pools")
kwargs = st.text_input("Filter JSON", placeholder=f'{{"author": "{get_username()}"}}', key="pool_kwargs")
try:
    kwargs = ast.literal_eval(kwargs)
except:
    if kwargs:
        st.warning("Invalid kwargs format. Please enter a valid dictionary.")
    kwargs = {}
pool_df = db.select_dataframe(model=Pool, limit=1000, **kwargs)
st.dataframe(pool_df)

st.write("#### Designs")
kwargs = st.text_input("Filter JSON", placeholder='{"name": "Personal project"}', key="design_kwargs")
try:
    kwargs = ast.literal_eval(kwargs)
except:
    if kwargs:
        st.warning("Invalid kwargs format. Please enter a valid dictionary.")
    kwargs = {}
design_df = db.select_dataframe(model=Design, limit=1000, **kwargs)
st.dataframe(design_df)

with st.expander("View design"):
    selected_design = st.selectbox("Select design", design_df.itertuples(), index=None, format_func=lambda x: x[0])
    if selected_design:
        design = db.select(model=Design, id=selected_design.Index)
        st.write(design)

st.write("#### Design jobs")
kwargs = st.text_input("Filter JSON", placeholder='{"job_result": True}', key="design_job_kwargs")
try:
    kwargs = ast.literal_eval(kwargs)
except:
    if kwargs:
        st.warning("Invalid kwargs format. Please enter a valid dictionary.")
    kwargs = {}
design_job_df = db.select_dataframe(model=DesignJob, limit=1000, **kwargs)
st.dataframe(design_job_df)

st.write("#### Descriptor jobs")
kwargs = st.text_input("Filter JSON", placeholder='{"job_result": True}', key="descriptor_job_kwargs")
try:
    kwargs = ast.literal_eval(kwargs)
except:
    if kwargs:
        st.warning("Invalid kwargs format. Please enter a valid dictionary.")
    kwargs = {}
descriptor_job_df = db.select_dataframe(model=DescriptorJob, limit=10000, **kwargs)
st.dataframe(descriptor_job_df)



import os
import re

import streamlit as st

from ovo import storage, get_username, config
from ovo.core.database.models import Pool, Design
from ovo.core.logic.descriptor_logic import submit_descriptor_workflow
from ovo.core.logic.round_logic import get_or_create_project_rounds
from ovo.core.database.models_proteinqc import ProteinQCWorkflow

from ovo import db
import zipfile


@st.dialog("Upload new pool of designs", width="large")
def create_new_pool():
    content = st.empty()
    with content.container():
        project_id = st.session_state.project.id

        rounds_by_id = get_or_create_project_rounds(project_id=project_id)
        round_ids = list(rounds_by_id.keys())
        round_id = st.selectbox(
            "Project Round",
            options=round_ids,
            format_func=lambda i: rounds_by_id[i].name,
            key="custom_pool_round_selectbox",
            index=len(round_ids) - 1,
            width=300,
        )

        files = st.file_uploader(
            "PDB files *",
            accept_multiple_files=True,
            type=["pdb", "zip"],
            key="uploader",
        )

        name = st.text_input("Pool name *", placeholder="Descriptive name for this collection of designs")

        if name and db.count(Pool, round_id=round_id, name=name):
            st.error(f"A pool with this name already exists in this round. Please choose a different name.")

        description = st.text_area(
            "Pool description",
            placeholder="Optional longer description of this pool",
        )

        with st.columns(2)[0]:
            chains = st.text_input(
                "Chain(s) to analyze",
                help="Chain IDs separated by comma (A,B,C), space (A B C) or concatenated (ABC)",
                value="",
            )
            chains = chains.replace(" ", "").replace(",", "")

        if not files:
            help = "No files selected"
        elif not name:
            help = "Please enter a pool name"
        elif not chains:
            help = "Please enter chain IDs to analyze"
        else:
            help = None
        with st.columns([3, 1])[1]:
            submit = st.button(
                "Upload pool",
                disabled=not files or not name or not chains,
                help=help,
                key="submit",
                type="primary",
                width="stretch",
            )

    if submit:
        assert re.match(r"^[A-Z]+$", chains), f"Invalid chains '{chains}'"
        content.empty()

        username = get_username()

        # Create pool
        pool = Pool(id=Pool.generate_id(), author=username, round_id=round_id, name=name, description=description)

        # Save the files to storage and create designs
        st.text("Uploading designs...")
        shared_args = dict(
            storage=storage,
            chains=list(chains),
            project_id=project_id,
            pool_id=pool.id,
        )
        designs = []
        for file in files:
            if file.name.endswith(".pdb"):
                designs.append(Design.from_pdb_file(filename=file.name, pdb_str=file.read().decode(), **shared_args))
            elif file.name.endswith(".zip"):
                # iterate over files in zip
                found = False
                with zipfile.ZipFile(file) as z:
                    for zip_info in z.infolist():
                        if zip_info.filename.startswith("__MACOSX/"):
                            continue
                        filename = os.path.basename(zip_info.filename)
                        if filename.startswith("."):
                            continue
                        if filename.endswith(".pdb"):
                            found = True
                            print("Reading", zip_info.filename)
                            pdb_str = z.read(zip_info.filename).decode()
                            designs.append(
                                Design.from_pdb_file(
                                    filename=filename,
                                    pdb_str=pdb_str,
                                    **shared_args,
                                )
                            )
                if not found:
                    raise ValueError(f"No PDB files found in zip archive '{file.name}'")

        if len(designs) > 1:
            st.text(f"Saving {len(designs):,} designs to DB...")

        # Save designs to db because we load them from db when submitting the proteinqc job
        db.save_all(designs + [pool])

        # Trigger sequence composition computation with local conda scheduler
        st.text("Submitting descriptor job...")
        submit_descriptor_workflow(
            workflow=ProteinQCWorkflow(
                tools=["seq_composition"],
                chains=list(chains),
                design_ids=[design.id for design in designs],
            ),
            scheduler_key=config.local_scheduler,
            project_id=project_id,
        )

        st.session_state.files = None
        st.text("✅ Done")

        st.rerun()



import os
import shutil
import tempfile
import zipfile

import streamlit as st
from humanize import naturalsize
from streamlit import session_state

from ovo import db
from ovo.app.utils.page_init import initialize_page
from ovo.core.database.models import Project
from ovo.core.logic.import_export_logic import export_project, import_project, export_import_project
from ovo import storage
from ovo.core.utils.formatting import safe_filename


EXPORT_DESIGNS_LIMIT = 20_000  # TODO should this be configured in config?


def show_counts(counts: dict[str, int]):
    # Display summary in columns
    col1, col2, col3 = st.columns(3)

    with col1:
        st.metric("Projects", counts.get("project"))
        st.metric("Rounds", counts.get("round"))
        st.metric("Pools", counts.get("pool"))

    with col2:
        st.metric("Total Designs", counts.get("design"))
        st.metric("Storage Files", counts.get("storage_file"))
        st.metric("Artifacts", counts.get("project_artifact"))

    with col3:
        st.metric("Descriptor Jobs", counts.get("descriptor_job"))
        st.metric("Descriptor Values", counts.get("descriptor_value"))


def export_tab():
    st.subheader("Export Project Data")
    st.markdown("""
    Export a complete project including all associated data to a ZIP file containing
    an SQLite database and all associated storage files.
    """)

    # Get all projects for the current user
    projects = db.select(Project, order_by="-created_date_utc")

    if not projects:
        st.warning("No projects found. Create a project first to export data.")
        return

    # Project selection
    st.subheader("Select Project to Export")
    projects_by_id = {project.id: project for project in projects}
    project_ids = sorted(projects_by_id.keys(), key=lambda p: projects_by_id[p].created_date_utc, reverse=True)
    project_id = st.selectbox(
        "Choose a project:",
        options=project_ids,
        index=project_ids.index(session_state.project.id)
        if session_state.project and session_state.project.id in project_ids
        else None,
        key="export_project_selector",
        format_func=lambda project_id: f"{projects_by_id[project_id].name} - {projects_by_id[project_id].author}",
    )

    if not project_id:
        return

    project = projects_by_id[project_id]

    # Show project information
    col1, col2 = st.columns(2)

    with col1:
        st.metric("Project Name", project.name)
        st.metric("Author", project.author)

    with col2:
        st.metric("Created", project.created_date_utc.strftime("%Y-%m-%d %H:%M UTC"))
        st.metric("Visibility", "Public" if project.public else "Private")

    selection_options = {
        "all": "All designs",
        "accepted": "Accepted designs",
    }
    selection = st.radio(
        "Export", options=selection_options.keys(), format_func=selection_options.get, key="export_selection"
    )
    accepted_only = selection == "accepted"

    # Get export summary
    st.subheader("Export Summary")
    with st.spinner("Preparing export..."):
        counts = export_import_project(
            source_db=db,
            source_storage=storage,
            dest_db=None,
            dest_dir=None,
            project_id=project_id,
            count_only=True,
            accepted_only=accepted_only,
        )

    st.markdown("The following data will be exported:")

    show_counts(counts)

    if counts["design"] > EXPORT_DESIGNS_LIMIT:
        # Do not allow exporting big projects for now due to memory limits
        if not accepted_only:
            st.error(
                f"Cannot export more than {EXPORT_DESIGNS_LIMIT} designs in the UI, "
                f'please select "Accepted designs" only, or use the commandline to export all designs.'
            )
        else:
            st.error(
                f"Cannot export more than {EXPORT_DESIGNS_LIMIT} designs in the UI, please use the commandline instead"
            )
        return

    # Export button
    if st.button("🚀 Start Export", type="primary", key="export_btn"):
        # Show progress
        progress_bar = st.progress(0)
        status_text = st.empty()

        status_text.text("Preparing export...")
        progress_bar.progress(10)

        # Perform the export
        status_text.text("Exporting project data...")
        progress_bar.progress(30)
        # TODO show actual progress

        zip_path = export_project(project_id, accepted_only=accepted_only)

        # Read the zip archive
        progress_bar.progress(80)
        status_text.text("Preparing download...")
        with open(zip_path, "rb") as f:
            zip_data = f.read()

        # Clean up temp file
        os.remove(zip_path)

        # Prepare download buttons
        progress_bar.progress(100)
        # clear status
        progress_bar.empty()
        status_text.empty()

        st.success("✅ Export completed successfully!")

        # Download section
        st.subheader("Download Export")

        st.download_button(
            label=f"📦 Download Project Export ({naturalsize(len(zip_data))} ZIP)",
            data=zip_data,
            file_name=f"{safe_filename(project.name)}_ovo_export.zip",
            mime="application/zip",
            help="Contains ovo.db database, storage files with relative paths, and config.yml template",
        )

        st.info(f"ZIP file size: {len(zip_data) / (1024 * 1024):.1f} MB")


def import_tab():
    st.subheader("Import Project Data")
    st.markdown("""
        Import project data from a ZIP file containing ovo.db and storage files.
        The ZIP file should be in the format created by the Export function.
        """)

    # Step 1: File uploader
    uploaded_file = st.file_uploader(
        "Choose a ZIP file to import",
        type="zip",
        help="Upload a ZIP file exported from OVO containing ovo.db and storage files",
        key="import_file_uploader",
    )

    if uploaded_file:
        with st.spinner("Importing ZIP file..."):
            temp_root = tempfile.mkdtemp(prefix="ovo")
            with zipfile.ZipFile(uploaded_file, "r") as zipf:
                zipf.extractall(temp_root)
            paths = os.listdir(temp_root)
            if not len(paths) == 1 or not os.path.isdir(os.path.join(temp_root, paths[0])):
                raise ValueError(f"Invalid ZIP file structure: expected a single root directory, got: {paths}")
            st.session_state.import_temp_dir = os.path.join(temp_root, paths[0])
            st.session_state.import_counts = import_project(st.session_state.import_temp_dir, count_only=True)
            st.session_state.import_done = False

    if not st.session_state.get("import_counts"):
        # Instructions
        st.subheader("Import Instructions")
        st.markdown("""
            **Import Requirements:**

            1. ZIP file must contain `ovo.db` (SQLite database file)
            2. ZIP file may contain `storage/` directory with data files (relative paths)
            3. ZIP file may contain `config.yml` (configuration template)
            4. ZIP file should be created using OVO's Export function

            **Import Process:**

            1. **Upload**: Choose a ZIP file to upload
            2. **Preview**: Click "Preview Project" to analyze the file and show project details
            3. **Customize**: Modify project name, author, and visibility settings
            4. **Import**: Click "Start Import" to import with your custom settings

            **Notes:**

            - Data is imported additively (existing data is preserved)
            - Duplicate IDs are skipped with warnings
            - Storage files are added with relative paths (no migration needed)
            - Configuration file is included as reference but not applied
            """)
        return

    st.subheader("Step 2: Project Details & Import Options")

    # Show original project information
    st.markdown("**Import Statistics:**")
    show_counts(st.session_state.import_counts)

    if st.button("Import data", type="primary", key="import_btn"):
        # Import the project
        with st.spinner("Importing..."):
            # TODO show progress
            import_project(st.session_state.import_temp_dir)

        st.session_state.import_done = True
        # Rerun to refresh projects list in sidebar, session will be cleared using logic below
        st.rerun()
    else:
        # Warning about import
        st.warning("""
        ⚠️ **Import Warning**

        - Importing will add new data to your current database
        - Storage files will be added to your storage system
        - This operation cannot be easily undone
        """)


initialize_page(page_title="Import & Export")

st.title("📦 Import & Export Project Data")

if st.session_state.get("import_done"):
    # Success message
    st.success("✅ Import completed successfully!")

    # Show import summary
    st.subheader("Import Summary")
    show_counts(st.session_state.import_counts)

    # Clear temp directory and session state
    shutil.rmtree(st.session_state.import_temp_dir)
    del st.session_state.import_temp_dir
    del st.session_state.import_counts
    del st.session_state.import_done

    st.stop()

# Create tabs for Import and Export
tab1, tab2 = st.tabs(["📤 Export", "📥 Import"])

with tab1:
    export_tab()

with tab2:
    import_tab()



from ovo.core.database import descriptors_rfdiffusion
from ovo.core.database.descriptors_refolding import AF2_PRIMARY_DESCRIPTORS, ESMFOLD_DESCRIPTORS
from ovo.core.database.models import NumericGlobalDescriptor, Descriptor

# SEQUENCE COMPOSITION

LENGTH = NumericGlobalDescriptor(
    name="Sequence length",
    description="Number of amino acids",
    tool="Sequence composition",
    key="proteinqc|seq_composition|length",
    comparison="does_not_apply",
)

# Warning value: 1.5 times the natural abundance
# Error value: 2 times the natural abundance

A_PERC = NumericGlobalDescriptor(
    name="Ala %",
    description="Percentage of alanine residues. Natural abundance in proteins: 8.76%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|A_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=13.14,
    error_value=17.52,
)

C_PERC = NumericGlobalDescriptor(
    name="Cys %",
    description="Percentage of cysteine residues. Natural abundance in proteins: 1.38%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|C_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=2.07,
    error_value=2.76,
)

D_PERC = NumericGlobalDescriptor(
    name="Asp %",
    description="Percentage of aspartic acid residues. Natural abundance in proteins: 5.49%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|D_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=8.234999999999999,
    error_value=10.98,
)

E_PERC = NumericGlobalDescriptor(
    name="Glu %",
    description="Percentage of glutamic acid residues. Natural abundance in proteins: 6.32%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|E_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=9.480000000000001,
    error_value=12.64,
)

F_PERC = NumericGlobalDescriptor(
    name="Phe %",
    description="Percentage of phenylalanine residues. Natural abundance in proteins: 3.87%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|F_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=5.805,
    error_value=7.74,
)

G_PERC = NumericGlobalDescriptor(
    name="Gly %",
    description="Percentage of glycine residues. Natural abundance in proteins: 7.03%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|G_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=10.545,
    error_value=14.06,
)

H_PERC = NumericGlobalDescriptor(
    name="His %",
    description="Percentage of histidine residues. Natural abundance in proteins: 2.26%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|H_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=3.39,
    error_value=4.52,
)

I_PERC = NumericGlobalDescriptor(
    name="Ile %",
    description="Percentage of isoleucine residues. Natural abundance in proteins: 5.49%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|I_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=8.234999999999999,
    error_value=10.98,
)

K_PERC = NumericGlobalDescriptor(
    name="Lys %",
    description="Percentage of lysine residues. Natural abundance in proteins: 5.19%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|K_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=7.785,
    error_value=10.38,
)

L_PERC = NumericGlobalDescriptor(
    name="Leu %",
    description="Percentage of leucine residues. Natural abundance in proteins: 9.68%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|L_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=14.52,
    error_value=19.36,
)

M_PERC = NumericGlobalDescriptor(
    name="Met %",
    description="Percentage of methionine residues. Natural abundance in proteins: 2.32%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|M_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=3.48,
    error_value=4.64,
)

N_PERC = NumericGlobalDescriptor(
    name="Asn %",
    description="Percentage of asparagine residues. Natural abundance in proteins: 3.93%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|N_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=5.895,
    error_value=7.86,
)

P_PERC = NumericGlobalDescriptor(
    name="Pro %",
    description="Percentage of proline residues. Natural abundance in proteins: 5.02%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|P_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=7.53,
    error_value=10.04,
)

Q_PERC = NumericGlobalDescriptor(
    name="Gln %",
    description="Percentage of glutamine residues. Natural abundance in proteins: 3.93%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|Q_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=5.895,
    error_value=7.86,
)

R_PERC = NumericGlobalDescriptor(
    name="Arg %",
    description="Percentage of arginine residues. Natural abundance in proteins: 5.78%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|R_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=8.67,
    error_value=11.56,
)

S_PERC = NumericGlobalDescriptor(
    name="Ser %",
    description="Percentage of serine residues. Natural abundance in proteins: 7.14%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|S_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=10.71,
    error_value=14.28,
)

T_PERC = NumericGlobalDescriptor(
    name="Thr %",
    description="Percentage of threonine residues. Natural abundance in proteins: 5.53%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|T_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=8.295,
    error_value=11.06,
)

V_PERC = NumericGlobalDescriptor(
    name="Val %",
    description="Percentage of valine residues. Natural abundance in proteins: 6.73%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|V_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=10.095,
    error_value=13.46,
)

W_PERC = NumericGlobalDescriptor(
    name="Trp %",
    description="Percentage of tryptophan residues. Natural abundance in proteins: 1.25%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|W_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=1.8750000000000003,
    error_value=2.5,
)

Y_PERC = NumericGlobalDescriptor(
    name="Tyr %",
    description="Percentage of tyrosine residues. Natural abundance in proteins: 2.91%.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|Y_perc",
    comparison="lower_is_better",  # High content of the same residue is suspicious
    min_value=0,
    max_value=100,
    warning_value=4.365,
    error_value=5.82,
)

NONPOLAR_PERC = NumericGlobalDescriptor(
    name="Nonpolar residues %",
    description="Percentage of nonpolar residues (G, A, V, L, I)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|nonpolar_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=90,
    error_value=100,
)

POLAR_UNCHARGED_PERC = NumericGlobalDescriptor(
    name="Polar uncharged residues %",
    description="Percentage of polar uncharged residues (S, T, C, N, Q, Y)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|polar_uncharged_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=90,
    error_value=100,
)

POSITIVE_PERC = NumericGlobalDescriptor(
    name="Positive residues %",
    description="Percentage of positively charged residues (K, R, H)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|positive_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=50,
    error_value=90,
)

NEGATIVE_PERC = NumericGlobalDescriptor(
    name="Negative residues %",
    description="Percentage of negatively charged residues (D, E)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|negative_perc",
    min_value=0,
    max_value=100,
    warning_value=50,
    error_value=90,
)

TURN_FORMING_PERC = NumericGlobalDescriptor(
    name="Turn forming residues %",
    description="Percentage of turn forming residues (N, G, P, S)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|turn_forming_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=50,
    error_value=90,
)

AROMATIC_PERC = NumericGlobalDescriptor(
    name="Aromatic residues %",
    description="Percentage of aromatic residues (F, Y, W)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|aromatic_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=50,
    error_value=90,
)

DENQ_PERC = NumericGlobalDescriptor(
    name="DENQ %",
    description="Percentage of aspartic acid (D), glutamic acid (E), asparagine (N), and glutamine (Q) residues",
    tool="Sequence composition",
    key="proteinqc|seq_composition|denq_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=50,
    error_value=90,
)

# Sequence composition - BIOPYTHON

AROMATICITY = NumericGlobalDescriptor(
    name="Aromaticity",
    description="Aromaticity value of a protein according to Lobry, 1994",
    tool="Sequence composition",
    key="proteinqc|seq_composition|aromaticity",
    comparison="lower_is_better",  # High aromaticity is associated with lower solubility, they are also expensive for biosynthesis so there is evolutionary pressure to reduce aromaticity, on the other hand they play important role in reactions and can stabilize the protein
    min_value=0,
    max_value=1,
    warning_value=0.2,
    error_value=0.5,
)

CHARGE_5_5 = NumericGlobalDescriptor(
    name="Charge at pH 5.5",
    description="Net charge of the protein sequence at pH 5.5",
    tool="Sequence composition",
    key="proteinqc|seq_composition|charge_5_5",
    comparison="does_not_apply",
)

CHARGE_7_4 = NumericGlobalDescriptor(
    name="Charge at pH 7.4",
    description="Net charge of the protein sequence at pH 7.4",
    tool="Sequence composition",
    key="proteinqc|seq_composition|charge_7_4",
    comparison="does_not_apply",
)

ISOELECTRIC_POINT = NumericGlobalDescriptor(
    name="Isoelectric point",
    description="pH at which protein carries no net eletrical charge",
    tool="Sequence composition",
    key="proteinqc|seq_composition|isoelectric_point",
    comparison="does_not_apply",
)  # FIXME: Maybe there should be a comparison

MEC_REDUCED = NumericGlobalDescriptor(
    name="MEC reduced",
    description="Molecular Extinction Coefficient (MEC) assuming cysteines (reduced)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|MEC_reduced",
    comparison="does_not_apply",
)

MEC_CYSTINES = NumericGlobalDescriptor(
    name="MEC cystines",
    description="Molecular Extinction Coefficient (MEC) assuming cystines residues (Cys-Cys-bond)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|MEC_cystines",
    comparison="does_not_apply",
)

HELIX_SEQ_PERC = NumericGlobalDescriptor(
    name="Helix-forming residues %",
    description="Percentage of residues which tend to be in Helix (V, I, Y, F, W, L)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|helix_perc",
    comparison="does_not_apply",
)

SHEET_SEQ_PERC = NumericGlobalDescriptor(
    name="Sheet-forming residues %",
    description="Percentage of residues which tend to be in Sheet (E, M, A, L)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|sheet_perc",
    comparison="does_not_apply",
)

TURN_SEQ_PERC = NumericGlobalDescriptor(
    name="Turn-forming residues %",
    description="Percentage of residues which tend to be in Turn (N, P, G, S)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|turn_perc",
    comparison="does_not_apply",
)

FLEXIBILITY_AVG = NumericGlobalDescriptor(
    name="Flexibility average",
    description="Average flexibility value of the protein sequence (according to Vihinen, 1994)",
    tool="Sequence composition",
    key="proteinqc|seq_composition|flexibility_avg",
    comparison="does_not_apply",
)

GRAVY = NumericGlobalDescriptor(
    name="GRAVY",
    description="Grand average of hydropathicity (GRAVY) value of the protein sequence (according to Kyte and Doolittle). Positive value indicates hydrophobic protein while negative values indicate hydrophilic protein.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|gravy",
    comparison="does_not_apply",
)

INSTABILITY_INDEX = NumericGlobalDescriptor(
    name="Instability index",
    description="Instability index of the protein sequence (according to Guruprasad et al., 1990). Any value above 40 means the protein is unstable (has a short half life).",
    tool="Sequence composition",
    key="proteinqc|seq_composition|instability_index",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
    warning_value=40,
    error_value=50,
)

MOLECULEAR_WEIGHT = NumericGlobalDescriptor(
    name="Molecular weight",
    description="Molecular weight of the protein sequence",
    tool="Sequence composition",
    key="proteinqc|seq_composition|molecular_weight",
    comparison="does_not_apply",
)

ENTROPY = NumericGlobalDescriptor(
    name="Sequence entropy",
    description="Diversity/orderedness of amino acids in a protein sequence; high values = more variety, low values = more compositional bias. For each overlapping 21-residue window (if sequence is longer than 21), the binary Shannon entropy of the amino acid composition is calculated and averaged to produce a single metric for the sequence. A sequence of identical amino acids has entropy 0, while a sequence with a uniform distribution of all 20 amino acids in each segment reaches log2(20) ~ 4.32.",
    tool="Sequence composition",
    key="proteinqc|seq_composition|avg_entropy",
    comparison="higher_is_better",
    min_value=2.976720519480519,
    max_value=3.580661971830986,
    warning_value=3.2,
    error_value=3,
)

# ESM-1v

ESM1V = NumericGlobalDescriptor(
    name="ESM-1v likelihood",
    description="Likelihood of native amino acid sequence based on pLM trained on large number of sequences",
    tool="ESM-1v",
    key="proteinqc|esm_1v|native_seq_avg_softmax",
    comparison="higher_is_better",
    min_value=0,
    max_value=1,
    warning_value=0.7,
    error_value=0.5,
)

# ESM-IF

ESMIF = NumericGlobalDescriptor(
    name="ESM-IF likelihood",
    description="Likelihood of native amino acid sequence based on the input structure",
    tool="ESM-IF",
    key="proteinqc|esm_if|native_seq_avg_softmax",
    comparison="higher_is_better",
    min_value=0,
    max_value=1,
    warning_value=0.3,
    error_value=0.2,
)

ESM_IF_NATIVE_SEQ_AVG_LOGITS = NumericGlobalDescriptor(
    name="ESM-IF native sequence average logits",
    description="Average logits for the native amino acid sequence based on the input structure",
    tool="ESM-IF",
    key="proteinqc|esm_if|native_seq_avg_logits",
    comparison="does_not_apply",
)

ESM_IF_NATIVE_SEQ_SUM_LOGITS = NumericGlobalDescriptor(
    name="ESM-IF native sequence sum logits",
    description="Sum of logits for the native amino acid sequence based on the input structure",
    tool="ESM-IF",
    key="proteinqc|esm_if|native_seq_sum_logits",
    comparison="does_not_apply",
)
# PROTEIN-SOL

PERC_SOLUBILITY = NumericGlobalDescriptor(
    name="Sequence solubility (percentage)",
    description="Predicted solubility from the sequence. Percentage values can exceed 100%",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|percent-sol",
    comparison="does_not_apply",
)

SOLUBILITY = NumericGlobalDescriptor(
    name="Sequence solubility (scaled)",
    description="Predicted solubility from the sequence (range: 0-1). Any scaled solubility value greater than 0.45 is predicted to have a higher solubility than the average soluble E.coli protein from the experimental solubility dataset Niwa et al 2009, and any protein with a lower scaled solubility value is predicted to be less soluble.",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|scaled-sol",
    comparison="higher_is_better",
    min_value=0,
    max_value=1,
    warning_value=0.45,
    error_value=0.3,
)

AVG_CHARGE = NumericGlobalDescriptor(
    name="Sequence charge",
    description="Average charge of the protein sequence",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|avg_charge",
    comparison="does_not_apply",
)

AVG_CHARGE_NEGATIVE = NumericGlobalDescriptor(
    name="Negative residue charge",
    description="Average charge of the negatively charged residues",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|negative_avg",
    comparison="does_not_apply",
)

AVG_CHARGE_POSITIVE = NumericGlobalDescriptor(
    name="Positive residue charge",
    description="Average charge of the positively charged residues",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|positive_avg",
    comparison="does_not_apply",
)

PERC_NEGATIVE = NumericGlobalDescriptor(
    name="Percentage of negative residues",
    description="Percentage of negative residues in the protein",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|perc_negative",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

PERC_POSITIVE = NumericGlobalDescriptor(
    name="Percentage of positive residues",
    description="Percentage of positive residues in the protein",
    tool="Protein-Sol",
    key="proteinqc|proteinsol|perc_positive",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

# DSSP

HELIX_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP Helix %",
    description="% of all amino acids in helices",
    tool="DSSP",
    key="proteinqc|dssp|% H",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

SHEET_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP Beta Sheet %",
    description="% of all amino acids in beta sheets",
    tool="DSSP",
    key="proteinqc|dssp|% E",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

TURN_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP Turn %",
    description="% of all amino acids in turns",
    tool="DSSP",
    key="proteinqc|dssp|% T",
    comparison="lower_is_better",  # Turns are most difficult for folding -> slow folding -> expression problems
    min_value=0,
    max_value=100,
    warning_value=20,
    error_value=50,
)

ISOLATED_BETA_BRIDGE_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP Isolated beta bridge residue %",
    description="% of all amino acids identified as isolated beta bridge residues",
    tool="DSSP",
    key="proteinqc|dssp|% B",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

HELIX_3_10_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP 3-10 helix %",
    description="% of all amino acids in 3-10 helices",
    tool="DSSP",
    key="proteinqc|dssp|% G",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

PI_HELIX_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP Pi helix %",
    description="% of all amino acids in pi helices",
    tool="DSSP",
    key="proteinqc|dssp|% I",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

BEND_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP Bend %",
    description="% of all amino acids in bends",
    tool="DSSP",
    key="proteinqc|dssp|% S",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

NONE_PERCENTAGE = NumericGlobalDescriptor(
    name="DSSP No secondary structure residues %",
    description="% of all amino acids in no secondary structure",
    tool="DSSP",
    key="proteinqc|dssp|% -",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

# Shape descriptors from MDAnalysis

ASPHERICITY = NumericGlobalDescriptor(
    name="Asphericity",
    description="Asphericity of the protein structure, value of 0 corresponds to a perfect sphere, values closer to 1 correspond to elongated proteins.",
    tool="MDAnalysis",
    key="proteinqc|dssp|asphericity",
    comparison="lower_is_better",
    min_value=0,
    warning_value=0.4,
    error_value=0.9,
    max_value=1,
)

SHAPE_PARAMETER = NumericGlobalDescriptor(
    name="Shape parameter",
    description="Shape parameter of the protein structure, S > 0 corresponds to prolate and S < 0 represents oblate",
    tool="MDAnalysis",
    key="proteinqc|dssp|shape_parameter",
    comparison="does_not_apply",
)

RADIUS_OF_GYRATION_ALL_ATOM = NumericGlobalDescriptor(
    name="Radius of gyration (all atom)",
    description="Radius of gyration of the protein structure",
    tool="MDAnalysis",
    key="proteinqc|dssp|radius_of_gyration",
    comparison="does_not_apply",
    min_value=0,
)

RADIUS_OF_GYRATION_NORM = NumericGlobalDescriptor(
    name="Radius of gyration normalized",
    description="Radius of gyration normalized by radius of gyration of a perfect sphere.  Values < 1 correspond to tight packing, values > 1 correspond to loose packing.",
    tool="MDAnalysis",
    key="proteinqc|dssp|radius_of_gyration_normalized",
    comparison="lower_is_better",
    min_value=0,
)

RAMACHANDRAN_ALLOWED_PERCENTAGE = NumericGlobalDescriptor(
    name="Ramachandran allowed angles %",
    description="Percentage of phi and psi angles in allowed regions of the Ramachandran plot (capturing 90% data points) created from a reference set of 500 PDB files.",
    tool="MDAnalysis",
    key="proteinqc|dssp|ramachandran_allowed_perc",
    comparison="higher_is_better",
    min_value=0,
    max_value=100,
)

RAMACHANDRAN_MARGINALLY_ALLOWED_PERCENTAGE = NumericGlobalDescriptor(
    name="Ramachandran marginally allowed angles %",
    description="Percentage of phi and psi angles in marginally allowed regions of the Ramachandran plot (capturing 99% data points) created from a reference set of 500 PDB files.",
    tool="MDAnalysis",
    key="proteinqc|dssp|ramachandran_marginally_allowed_perc",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

RAMACHANDRAN_OUTLIERS_PERCENTAGE = NumericGlobalDescriptor(
    name="Ramachandran outliers angles %",
    description="Percentage of phi and psi angles in outlier regions of the Ramachandran plot created from a reference set of 500 PDB files.",
    tool="MDAnalysis",
    key="proteinqc|dssp|ramachandran_outliers_perc",
    comparison="lower_is_better",
    min_value=0,
    max_value=100,
)

# PEP-PATCH

# Hydrophobic
HYDROPHOBIC_TOTAL_AREA = NumericGlobalDescriptor(
    name="Hydrophobic patch area",
    description="Surface area predicted to be in a hydrophobic patch (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area,hydrophobic",
    comparison="does_not_apply",
)

HYDROPHOBIC_TOTAL_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized hydrophobic patch area",
    description="Surface area predicted to be in a hydrophobic patch, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_normalized,hydrophobic",
    comparison="lower_is_better",
    min_value=0,
    max_value=1,
    warning_value=0.25,
    error_value=0.45,
)

HYDROPHOBIC_TOP_PATCH_AREA = NumericGlobalDescriptor(
    name="Hydrophobic top1 patch area",
    description="Area of the largest hydrophobic patch on the protein surface",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area,hydrophobic",
    comparison="does_not_apply",
)

HYDROPHOBIC_TOP_PATCH_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized hydrophobic top1 patch area",
    description="Area of the largest hydrophobic patch on the protein surface, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_normalized,hydrophobic",
    comparison="does_not_apply",
)

HYDROPHOBIC_SAP_SCORE = NumericGlobalDescriptor(
    name="Hydrophobic SAP score",
    description="SAP (Spatial Aggregation Propensity) score",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sap_score,hydrophobic",
    comparison="does_not_apply",
)

HYDROPHOBIC_SASA = NumericGlobalDescriptor(
    name="Hydrophobic SASA",
    description="Solvent accessible surface area (SASA)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sasa,hydrophobic",
    comparison="does_not_apply",
)

HYDROPHOBIC_SURFACE_INTEGRAL = NumericGlobalDescriptor(
    name="Hydrophobicity potential",
    description="Integral of hydrophobicity values over solvent-accessible surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|total_surface_integral,hydrophobic",
    comparison="does_not_apply",
)

# Electrostatic pH 5.5
CHARGED_POSITIVE_5_5_TOTAL_AREA = NumericGlobalDescriptor(
    name="Positive patch area at pH 5.5",
    description="Surface area predicted to be in a positively charged patch at pH 5.5 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_+,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_POSITIVE_5_5_TOTAL_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized positive patch area at pH 5.5",
    description="Surface area predicted to be in a positively charged patch at pH 5.5, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_+_normalized,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_POSITIVE_5_5_TOP_PATCH_AREA = NumericGlobalDescriptor(
    name="Positive top1 patch area at pH 5.5",
    description="Area of the largest positively charged patch on the protein surface at pH 5.5 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_+,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_POSITIVE_5_5_TOP_PATCH_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized positive top1 patch area at pH 5.5",
    description="Area of the largest positively charged patch on the protein surface at pH 5.5, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_+_normalized,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_5_5_TOTAL_AREA = NumericGlobalDescriptor(
    name="Negative patch area at pH 5.5",
    description="Surface area predicted to be in a negatively charged patch at pH 5.5 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_-,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_5_5_TOTAL_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized negative patch area at pH 5.5",
    description="Surface area predicted to be in a negatively charged patch at pH 5.5, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_-_normalized,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_5_5_TOP_PATCH_AREA = NumericGlobalDescriptor(
    name="Negative top1 patch area at pH 5.5",
    description="Area of the largest negatively charged patch on the protein surface at pH 5.5 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_-,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_5_5_TOP_PATCH_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized negative top1 patch area at pH 5.5",
    description="Area of the largest negatively charged patch on the protein surface at pH 5.5, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_-_normalized,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_5_5 = NumericGlobalDescriptor(
    name="Electrostatic volume integral at pH 5.5",
    description="Integral of all electrostatic potential values over the solvent-accessible volume, overall measure of the electrostatic potential",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_HIGH_5_5 = NumericGlobalDescriptor(
    name="Positive electrostatic regions volume integral at pH 5.5",
    description="Integral of electrostatic potential values above a positive threshold, highlighting strongly positive regions",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_high,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_LOW_5_5 = NumericGlobalDescriptor(
    name="Negative electrostatic regions volume integral at pH 5.5",
    description="Integral of electrostatic potential values below a negative threshold, highlighting strongly negative regions",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_low,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_POSITIVE_5_5 = NumericGlobalDescriptor(
    name="Positive electrostatic volume integral at pH 5.5",
    description="Integral of all positive electrostatic potential values over the solvent-accessible volume",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_pos,electrostatic_5.5",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_NEGATIVE_5_5 = NumericGlobalDescriptor(
    name="Negative electrostatic volume integral at pH 5.5",
    description="Integral of all negative electrostatic potential values over the solvent-accessible volume",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_neg,electrostatic_5.5",
    comparison="does_not_apply",
)
# Electrostatic pH 7.4
CHARGED_POSITIVE_7_4_TOTAL_AREA = NumericGlobalDescriptor(
    name="Positive patch area at pH 7.4",
    description="Surface area predicted to be in a positively charged patch at pH 7.4 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_+,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_POSITIVE_7_4_TOTAL_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized positive patch area at pH 7.4",
    description="Surface area predicted to be in a positively charged patch at pH 7.4, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_+_normalized,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_POSITIVE_7_4_TOP_PATCH_AREA = NumericGlobalDescriptor(
    name="Positive top1 patch area at pH 7.4",
    description="Area of the largest positively charged patch on the protein surface at pH 7.4 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_+,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_POSITIVE_7_4_TOP_PATCH_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized positive top1 patch area at pH 7.4",
    description="Area of the largest positively charged patch on the protein surface at pH 7.4, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_+_normalized,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_7_4_TOTAL_AREA = NumericGlobalDescriptor(
    name="Negative patch area at pH 7.4",
    description="Surface area predicted to be in a negatively charged patch at pH 7.4 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_-,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_7_4_TOTAL_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized negative patch area at pH 7.4",
    description="Surface area predicted to be in a negatively charged patch at pH 7.4, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|sum_patch_area_-_normalized,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_7_4_TOP_PATCH_AREA = NumericGlobalDescriptor(
    name="Negative top1 patch area at pH 7.4",
    description="Area of the largest negatively charged patch on the protein surface at pH 7.4 (Å²)",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_-,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_NEGATIVE_7_4_TOP_PATCH_AREA_NORM = NumericGlobalDescriptor(
    name="Normalized negative top1 patch area at pH 7.4",
    description="Area of the largest negatively charged patch on the protein surface at pH 7.4, normalized by the total surface area",
    tool="PEP-Patch",
    key="proteinqc|peppatch|top1_patch_area_-_normalized,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_7_4 = NumericGlobalDescriptor(
    name="Electrostatic volume integral at pH 7.4",
    description="Integral of all electrostatic potential values over the solvent-accessible volume, overall measure of the electrostatic potential",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_HIGH_7_4 = NumericGlobalDescriptor(
    name="Positive electrostatic regions volume integral at pH 7.4",
    description="Integral of electrostatic potential values above a positive threshold, highlighting strongly positive regions",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_high,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_LOW_7_4 = NumericGlobalDescriptor(
    name="Negative electrostatic regions volume integral at pH 7.4",
    description="Integral of electrostatic potential values below a negative threshold, highlighting strongly negative regions",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_low,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_POSITIVE_7_4 = NumericGlobalDescriptor(
    name="Positive electrostatic volume integral at pH 7.4",
    description="Integral of all positive electrostatic potential values over the solvent-accessible volume",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_pos,electrostatic_7.4",
    comparison="does_not_apply",
)

CHARGED_INTEGRAL_NEGATIVE_7_4 = NumericGlobalDescriptor(
    name="Negative electrostatic volume integral at pH 7.4",
    description="Integral of all negative electrostatic potential values over the solvent-accessible volume",
    tool="PEP-Patch",
    key="proteinqc|peppatch|integral_neg,electrostatic_7.4",
    comparison="does_not_apply",
)
# Electrostatic pH 7.4
# Top descriptors
PROTEINQC_SEQUENCE_DESCRIPTORS = [
    LENGTH,
    SOLUBILITY,
    ISOELECTRIC_POINT,
    ESM1V,
    AROMATICITY,
    CHARGE_7_4,
    ENTROPY,
]
PROTEINQC_STRUCTURE_DESCRIPTORS = [
    ASPHERICITY,
    HYDROPHOBIC_TOTAL_AREA_NORM,
    CHARGED_POSITIVE_7_4_TOTAL_AREA_NORM,
    HELIX_PERCENTAGE,
    SHEET_PERCENTAGE,
    TURN_PERCENTAGE,
    ESMIF,
]

# All
SEQ_COMPOSITION_DESCRIPTORS = [
    LENGTH,
    A_PERC,
    C_PERC,
    D_PERC,
    E_PERC,
    F_PERC,
    G_PERC,
    H_PERC,
    I_PERC,
    K_PERC,
    L_PERC,
    M_PERC,
    N_PERC,
    P_PERC,
    Q_PERC,
    R_PERC,
    S_PERC,
    T_PERC,
    V_PERC,
    W_PERC,
    Y_PERC,
    NONPOLAR_PERC,
    POLAR_UNCHARGED_PERC,
    POSITIVE_PERC,
    NEGATIVE_PERC,
    TURN_FORMING_PERC,
    AROMATIC_PERC,
    DENQ_PERC,
    AROMATICITY,
    CHARGE_5_5,
    CHARGE_7_4,
    ISOELECTRIC_POINT,
    MEC_REDUCED,
    MEC_CYSTINES,
    HELIX_SEQ_PERC,
    SHEET_SEQ_PERC,
    TURN_SEQ_PERC,
    FLEXIBILITY_AVG,
    GRAVY,
    INSTABILITY_INDEX,
    MOLECULEAR_WEIGHT,
    ENTROPY,
]
ESM_DESCRIPTORS = [ESM1V, ESMIF, ESM_IF_NATIVE_SEQ_AVG_LOGITS, ESM_IF_NATIVE_SEQ_SUM_LOGITS]
PROTEIN_SOL_DESCRIPTORS = [
    SOLUBILITY,
    AVG_CHARGE,
    PERC_SOLUBILITY,
    AVG_CHARGE_NEGATIVE,
    AVG_CHARGE_POSITIVE,
    PERC_NEGATIVE,
    PERC_POSITIVE,
]
DSSP_DESCRIPTORS = [
    HELIX_PERCENTAGE,
    SHEET_PERCENTAGE,
    TURN_PERCENTAGE,
    ISOLATED_BETA_BRIDGE_PERCENTAGE,
    HELIX_3_10_PERCENTAGE,
    PI_HELIX_PERCENTAGE,
    BEND_PERCENTAGE,
    NONE_PERCENTAGE,
]
SHAPE_DESCRIPTORS = [
    ASPHERICITY,
    SHAPE_PARAMETER,
    RADIUS_OF_GYRATION_ALL_ATOM,
    RADIUS_OF_GYRATION_NORM,
    RAMACHANDRAN_ALLOWED_PERCENTAGE,
    RAMACHANDRAN_MARGINALLY_ALLOWED_PERCENTAGE,
    RAMACHANDRAN_OUTLIERS_PERCENTAGE,
]

PEP_PATCH_HYDROPHOBICITY_DESCRIPTORS = [
    HYDROPHOBIC_SAP_SCORE,
    HYDROPHOBIC_SASA,
    HYDROPHOBIC_TOTAL_AREA,
    HYDROPHOBIC_TOTAL_AREA_NORM,
    HYDROPHOBIC_TOP_PATCH_AREA,
    HYDROPHOBIC_TOP_PATCH_AREA_NORM,
    HYDROPHOBIC_SURFACE_INTEGRAL,
]
PEP_PATCH_HYDROPHOBICITY_DESCRIPTORS_BY_KEY = {d.key: d for d in PEP_PATCH_HYDROPHOBICITY_DESCRIPTORS}
PEP_PATCH_CHARGE_DESCRIPTORS = [
    CHARGED_POSITIVE_5_5_TOTAL_AREA,
    CHARGED_POSITIVE_5_5_TOTAL_AREA_NORM,
    CHARGED_POSITIVE_5_5_TOP_PATCH_AREA,
    CHARGED_POSITIVE_5_5_TOP_PATCH_AREA_NORM,
    CHARGED_NEGATIVE_5_5_TOTAL_AREA,
    CHARGED_NEGATIVE_5_5_TOTAL_AREA_NORM,
    CHARGED_NEGATIVE_5_5_TOP_PATCH_AREA,
    CHARGED_NEGATIVE_5_5_TOP_PATCH_AREA_NORM,
    CHARGED_INTEGRAL_5_5,
    CHARGED_INTEGRAL_HIGH_5_5,
    CHARGED_INTEGRAL_LOW_5_5,
    CHARGED_INTEGRAL_POSITIVE_5_5,
    CHARGED_INTEGRAL_NEGATIVE_5_5,
    CHARGED_POSITIVE_7_4_TOTAL_AREA,
    CHARGED_POSITIVE_7_4_TOTAL_AREA_NORM,
    CHARGED_POSITIVE_7_4_TOP_PATCH_AREA,
    CHARGED_POSITIVE_7_4_TOP_PATCH_AREA_NORM,
    CHARGED_NEGATIVE_7_4_TOTAL_AREA,
    CHARGED_NEGATIVE_7_4_TOTAL_AREA_NORM,
    CHARGED_NEGATIVE_7_4_TOP_PATCH_AREA,
    CHARGED_NEGATIVE_7_4_TOP_PATCH_AREA_NORM,
    CHARGED_INTEGRAL_7_4,
    CHARGED_INTEGRAL_HIGH_7_4,
    CHARGED_INTEGRAL_LOW_7_4,
    CHARGED_INTEGRAL_POSITIVE_7_4,
    CHARGED_INTEGRAL_NEGATIVE_7_4,
]
PEP_PATCH_CHARGE_DESCRIPTORS_BY_KEY = {d.key: d for d in PEP_PATCH_CHARGE_DESCRIPTORS}
PEP_PATCH_DESCRIPTORS = PEP_PATCH_HYDROPHOBICITY_DESCRIPTORS + PEP_PATCH_CHARGE_DESCRIPTORS

# New descriptors
DESCRIPTORS = [v for v in globals().values() if isinstance(v, Descriptor)]

# All descriptors shown on ProteinQC page - includes descriptors from other pipelines
PROTEINQC_MAIN_DESCRIPTORS = (
    PROTEINQC_SEQUENCE_DESCRIPTORS
    + PROTEINQC_STRUCTURE_DESCRIPTORS
    + AF2_PRIMARY_DESCRIPTORS
    + ESMFOLD_DESCRIPTORS
    + descriptors_rfdiffusion.BACKBONE_METRICS
)

PRESETS = [
    {
        "label": "Sequence pI & Pep-Patch Hydrophobic SAP Score",
        "x": ISOELECTRIC_POINT,
        "y": HYDROPHOBIC_SAP_SCORE,
    },
    {
        "label": "Sequence Hydrophobicity & Charge",
        "x": GRAVY,
        "y": CHARGE_7_4,
    },
    {
        "label": "Aromatic vs. Positive residues %",
        "x": AROMATIC_PERC,
        "y": POSITIVE_PERC,
    },
]



import fnmatch
import warnings
from typing import Optional, Union

import numpy as np
import pandas as pd


def get_feat_name(
    method_name,
    feat_name,
    cutoff: Union[float, tuple[float]] = None,
    sub_df_prefix: str = "",
    mab_type: str = "",
):
    """Return feature name using the template in this function.

    example
    H_CDR3_surface_integral_high_0.4_fv|electrostatic_7.4    # with sub_df_prefix `H_CDR3_`
    top5_patches_mean_area_-_fv|electrostatic_7.4_c[3, -3]   # with patch cutoff_str `c[3, -3]`
    """
    cutoff_str = ""
    # if cutoff:
    #     cutoff_str = f"_c{cutoff}"

    if mab_type:
        mab_type = f"_{mab_type}"

    return f"{sub_df_prefix}{feat_name}{mab_type},{method_name}{cutoff_str}"


def get_vertices_sub_dfs(vertices: pd.DataFrame):
    """Produces vertex dfs with names:

    'Total_' (all),
    'Chain_{chain_id}_' (chain vertices) for each chain_id present in the structure.
    """
    vs = vertices
    unique_chains = vertices.chain_id.unique()

    dfs = {f"chain_{chain_id}_": vs[vs.chain_id == chain_id] for chain_id in unique_chains}
    dfs["total_"] = vs
    return dfs.items()


def expand_aggrs(filtered_df, col_pattern_aggrs: dict[str, list]):
    """Expanding the unix-style wildcards (using fnmatch) to actual column names."""

    col_aggrs = {}
    for col_pattern, aggrs in col_pattern_aggrs.items():
        # Use fnmatch to find matching columns
        matching_cols = [col for col in filtered_df.columns if fnmatch.fnmatch(col, col_pattern)]
        if not matching_cols:
            warnings.warn(
                f"No columns matched the colname pattern `{col_pattern}` - did you specify the pattern correctly?"
            )
        for col in matching_cols:
            col_aggrs[col] = aggrs
    return col_aggrs


def aggregate_patch_data(patches_df, top_ns=(1, 2, 3, 5)):
    """Create structure-level features (fixed length) from the dataframe of patches (variable length).

    Three types of features are returned:
    - over all patches (total path area, patch intensity etc)
    - over top n (1,2,...) patches
        - the values for n-largest patch (area, intensity,...)
        - values if 1..n-largest patch aggregated (equal weight for each patch disregarding its size)
            - e.g. for top3, the values for the largest three patches are meaned/summed, for five, the largest 5
                patches' values are meaned

    """

    feats = {
        "num_patches": len(patches_df),
    }

    col_aggrs = {
        "area": ["mean", "sum"],
        "intensity_sum": ["mean", "sum"],
        "intensity_mean": ["mean"],
        "mean_buriedness_*": ["mean"],
        "min_buriedness_*": [],
        "max_buriedness_*": [],
    }

    col_aggrs = expand_aggrs(patches_df, col_aggrs)

    fields = list(col_aggrs.keys())

    # struct-level aggregation, all patches
    for field in fields:
        for aggr in col_aggrs[field]:
            feats[f"{aggr}_patch_{field}"] = patches_df[field].aggregate(aggr)

    # 1) Values for top-n largest patch
    # 2) Aggregated values for all patches 1..top-n
    for n in top_ns:
        # if there are less than n patches - nans
        nlargest_patches = (
            patches_df.nlargest(n, "area")
            if len(patches_df) >= n
            else pd.DataFrame({c: [np.nan] * n for c in patches_df.columns})
        )

        top_n_patch = nlargest_patches.iloc[-1]

        for field in fields:
            feats[f"top{n}_patch_{field}"] = top_n_patch[field]

            if n == 1:  # no aggr. needed
                continue

            for aggr in col_aggrs[field]:
                top_n_feat_aggr = nlargest_patches[field].aggregate(aggr)
                feats[f"top{n}_patches_{aggr}_{field}"] = top_n_feat_aggr

    return feats


def compute_surface_integrals(
    vertices: pd.DataFrame, high_cutoffs: list = (), low_cutoffs: list = ()
) -> dict[str, float]:
    """Computes surface integrals of the potential over all vertices. sum_vertices(potential value * vertex area)

    Additionally, high and low integrals -> the summed quantity, value, is only taken as what is above/below a cutoff threshold.
    """
    features = {}

    features["surface_integral"] = np.sum(vertices["value"] * vertices["area"])

    for cutoff in high_cutoffs:
        features[f"surface_integral_high_{cutoff}"] = np.sum(
            np.maximum(vertices["value"] - cutoff, 0) * vertices["area"]
        )

    for cutoff in low_cutoffs:
        features[f"surface_integral_low_{cutoff}"] = np.sum(
            np.minimum(vertices["value"] - cutoff, 0) * vertices["area"]
        )

    return features


def get_buriedness_for_patches(vertices: pd.DataFrame, atom_buriedness: pd.DataFrame) -> pd.DataFrame:
    """Compute buriedness scores of the patches.

    Buriedness := #atoms in a set radius from query atom.
    Mean -> mean the buriedness of the patch's atoms
    Max, min -> take max/min of the patch's atoms, of their buriedness

    Returns a DataFrame with patch_id and buriedness columns.
    """
    # Select only vertices in patches
    vertices = vertices[vertices.patch_id != 0]
    # Merge with atom_buriedness on 'atom'
    vertices = vertices.drop(columns="chain_id").merge(
        atom_buriedness, left_on="atom", right_index=True
    )  # drop the col because atom_buriedness already has it

    by_patch = vertices.groupby("patch_id")

    # Calculate the total area of each patch
    total_area = by_patch["area"].sum()

    # Calculate the weighted mean, max, and min buriedness for each patch (for each buriedness radius)
    patch_buriedness_df = pd.DataFrame(index=by_patch.size().index)
    for radius_col in atom_buriedness.columns:
        if not radius_col.startswith("buriedness_"):
            continue

        weighted_sum_of_buriedness = (vertices[radius_col] * vertices["area"]).groupby(vertices["patch_id"]).sum()
        patch_buriedness_df[f"mean_{radius_col}"] = weighted_sum_of_buriedness / total_area
        patch_buriedness_df[f"max_{radius_col}"] = by_patch[radius_col].max()
        patch_buriedness_df[f"min_{radius_col}"] = by_patch[radius_col].min()

    return patch_buriedness_df


def get_patch_df(patch_vertices, atom_buriedness: Optional[pd.DataFrame]):
    """Aggregate values for patches' vertices into patch-level values.

    Add buriedness scores to each patch, if atom_buriedness is not None.
    Returns a DataFrame with patch_id and the aggregated values/buriedness score.
    """
    patch_df = patch_vertices.groupby("patch_id").aggregate(
        intensity_mean=pd.NamedAgg("value", "mean"),
        intensity_sum=pd.NamedAgg("value", "sum"),
        area=pd.NamedAgg("area", "sum"),
    )

    if atom_buriedness is not None:
        buriedness_df = get_buriedness_for_patches(patch_vertices, atom_buriedness)
        patch_df = patch_df.merge(buriedness_df, on="patch_id")

    return patch_df


def get_patch_area_features(patch_vertices):
    """Return dict of patch area broken down per all/selected chain/CDR1/2/3 vertices."""
    feats = {}
    # include total patch area per all/selected chain/CDR1/2/3
    for sub_df_name, sub_df in get_vertices_sub_dfs(patch_vertices):
        feats[f"{sub_df_name}patch_area"] = sub_df[sub_df["patch_id"] != 0]["area"].sum()
    return feats


def get_global_electrostatic_descriptors_for_patches(
    patch_vertices: pd.DataFrame, atom_buriedness: Optional[pd.DataFrame]
) -> dict[str, float]:
    """Compute scalar structure-level features from patch vertices.

    Returns a dict of scalar features {name: value}.
    - features with _+ in their name for positive patches, _- for negative

    Example feature names (when additionally passed through get_feat_name):
    - sum_patch_area_+|electrostatic_7.4_c[2, -2]
    - top5_patch_area_-|electrostatic_7.4_c[2, -2]       # this is value for the 5th largest patch
    - top5_patches_sum_area_-|electrostatic_7.4_c[2, -2] # this is value summed over _all_ 5 largest patches

    (Features can be nan – e.g. there are less then 5 patches but we're computing a value for top-5 patch.)
    """

    out = {}
    positive_df = patch_vertices[patch_vertices.patch_id > 0]
    negative_df = patch_vertices[patch_vertices.patch_id < 0]

    positive_patch_feats = aggregate_patch_data(get_patch_df(positive_df, atom_buriedness))
    negative_patch_feats = aggregate_patch_data(get_patch_df(negative_df, atom_buriedness))

    def with_suffix(dictionary, key_suffix):
        return {f"{k}{key_suffix}": v for k, v in dictionary.items()}

    out.update(with_suffix(positive_patch_feats, "_+"))
    out.update(with_suffix(negative_patch_feats, "_-"))
    out.update(with_suffix(get_patch_area_features(positive_df), "_+"))
    out.update(with_suffix(get_patch_area_features(negative_df), "_-"))
    return out


def get_global_hydrophobic_descriptors_for_patches(
    patch_vertices: pd.DataFrame, atom_buriedness: pd.DataFrame
) -> dict[str, float]:
    """Compute scalar structure-level features from patch vertices.

    Returns a dict of scalar features {name: value} like:
    - mean_patch_area|hydrophobic_Ja_c0.8
    - top2_patch_intensity_sum|hydrophobic_Ja_c0.8          # this is value for the 2nd largest patch
    - top2_patches_mean_intensity_sum|hydrophobic_Ja_c0.8   # this is value meaned over _both_ 2 largest patches

    (Names when additionally passed through get_feat_name)
    (Features can be nan – e.g. there are less then 5 patches but we're computing a value for top-5 patch.)
    """
    feats = aggregate_patch_data(get_patch_df(patch_vertices[patch_vertices.patch_id > 0], atom_buriedness))
    # include total patch area per all/selected chain/CDR1/2/3
    feats.update(get_patch_area_features(patch_vertices))
    return feats


def get_global_hydrophobic_descriptors(peppatch_hydrophobic_output: dict[str, np.ndarray]):
    """Aggregate per-atom descriptors computed by PEP-Patch hydrophobic; over the whole structure.

    (these outputs are enabled with extra flags to PEP-Patch)"""
    output = peppatch_hydrophobic_output

    # choose trajectory frame 0 because there may be multiple structures (PEP-Patch accepts mol. dynamics trajectory),
    # we have only a single structure
    FRAME = 0
    return dict(
        # sasa_score is similar to the potential integral, however this uses the raw hydrophobicity scale values
        # (but using which the potential is calculated :)
        sasa_score=np.sum(output["surfscore"][FRAME]),
        positive_sasa_score=np.sum(np.maximum(output["surfscore"][FRAME], 0)),
        sap_score=np.sum(output["sap"][FRAME]),
        surrounding_hydrophobicity=np.sum(output["surrounding_hydrophobicity"][FRAME]),
        sasa=np.sum(output["saa"][FRAME]),
    )



import subprocess
import os
import pandas as pd
import pytest

from ovo.core.utils.resources import RESOURCES_DIR
from ovo.core.utils.tests import get_test_scheduler, prepare_test_output_dir, assert_similar_sequence


def test_5eli_binder_fastrelax():
    scheduler = get_test_scheduler()
    output_dir = prepare_test_output_dir(__file__)
    scheduler.run(
        "rfdiffusion-end-to-end",
        output_dir=output_dir,
        params=dict(
            max_memory="4G",
            batch_size="2",
            design_type="binder",
            rfdiffusion_input_pdb=RESOURCES_DIR / "examples/inputs/5ELI_A.pdb",
            rfdiffusion_num_designs=1,
            rfdiffusion_contig="A74-97/0 20",
            rfdiffusion_run_parameters="diffuser.T=15 inference.deterministic=True",
            mpnn_run_parameters="-seed 42",
            mpnn_fastrelax_cycles=1,
            refolding_tests="af2_model_1_multimer_tt_3rec",
        ),
    )
    print("Saved to:", output_dir)
    subprocess.run(["ls", "-lR", output_dir.rstrip("/") + "/"], check=True)

    seq_composition = pd.read_csv(os.path.join(output_dir, "contig1_batch1/seq_composition.csv"))
    assert len(seq_composition) == 2
    first_seq = seq_composition.iloc[0]
    assert_similar_sequence(first_seq.sequence, "TLTLTITTTNGVTTITVTVS", 0.2)
    second_seq = seq_composition.iloc[1]
    assert_similar_sequence(second_seq.sequence, "TTTITILLKDGKVTVTVTDV", 0.2)

    initial_guess = pd.read_json(
        os.path.join(output_dir, "contig1_batch1/af2_model_1_multimer_tt_3rec.jsonl"), lines=True
    )
    assert len(initial_guess) == 2
    first_scores = initial_guess.iloc[0]
    assert first_scores["target_aligned_binder_rmsd"] < 100
    assert first_scores["binder_plddt"] > 5.0
    second_scores = initial_guess.iloc[1]
    assert second_scores["target_aligned_binder_rmsd"] < 100
    assert second_scores["binder_plddt"] > 5.0



import os

import streamlit as st

from ovo import db, storage
from ovo.app.components.custom_elements import wrapped_columns
from ovo.app.components.molstar_custom_component.dataclasses import ContigSegment
from ovo.core.database import (
    Design,
    descriptors,
    StructureFileDescriptor,
    descriptors_rfdiffusion,
    descriptors_refolding,
    descriptors_proteinqc,
    descriptors_bindcraft,
)
from ovo.core.database.models_rfdiffusion import (
    RFdiffusionWorkflow,
    RFdiffusionBinderDesignWorkflow,
    RFdiffusionScaffoldDesignWorkflow,
)
from ovo.core.database.descriptors import (
    ALL_DESCRIPTORS_BY_KEY,
    SEQUENCE_DESIGN_PATH_DESCRIPTORS,
    STRUCTURE_PATH_DESCRIPTORS,
)

from ovo.app.components.molstar_custom_component import (
    molstar_custom_component,
    StructureVisualization,
    ChainVisualization,
    ContigsParser,
)
from ovo.core.utils.pdb import align_multiple_proteins_pdb, pdb_to_mmcif, filter_pdb_str, get_sequences_from_pdb_str
from ovo.core.utils.colors import get_color_from_str
from ovo.app.utils.cached_db import (
    get_cached_design_descriptors,
    get_cached_design,
    get_cached_pool,
    get_cached_design_job,
    get_cached_designs,
)
from ovo.core.utils.residue_selection import from_segments_to_hotspots


def show_design_metrics(design_id: str, descriptor_keys: list[str]):
    descriptor_values = get_cached_design_descriptors(design_id, descriptor_keys=descriptor_keys)
    columns = wrapped_columns(len(descriptor_values), wrap=4)
    for column, (descriptor_key, value) in zip(columns, descriptor_values.items()):
        descriptor = ALL_DESCRIPTORS_BY_KEY[descriptor_key]
        column.metric(
            label=descriptor.name,
            help=descriptor.description,
            value=descriptor.format(value),
        )


def select_structure_prediction_descriptor(paths: dict[str, str]) -> StructureFileDescriptor | None:
    prediction_descriptors = [
        d for d in STRUCTURE_PATH_DESCRIPTORS if d.structure_type == "prediction" and d.key in paths
    ]

    if not prediction_descriptors:
        st.info("No structure prediction was run for this design.")
        return None
    elif len(prediction_descriptors) == 1:
        prediction_descriptor = prediction_descriptors[0]
    else:
        prediction_descriptor = st.segmented_control(
            "Structure prediction method",
            options=prediction_descriptors,
            key="structure_prediction_method",
            format_func=lambda d: d.name,
        )
    return prediction_descriptor


def rfdiffusion_scaffold_design_visualization(design_id: str | None):
    design: Design = get_cached_design(design_id)

    chain_contigs = [c.contig + "/0" for c in design.spec.chains]
    st.write(f"Contig: **{' '.join(chain_contigs)}**")

    show_design_metrics(
        design_id,
        descriptor_keys=[
            descriptors_proteinqc.LENGTH.key,
            descriptors_rfdiffusion.RADIUS_OF_GYRATION.key,
            descriptors_rfdiffusion.PYDSSP_HELIX_PERCENT.key,
            descriptors_rfdiffusion.PYDSSP_SHEET_PERCENT.key,
            descriptors_refolding.AF2_PRIMARY_PAE.key,
            descriptors_refolding.AF2_PRIMARY_DESIGN_RMSD.key,
            descriptors_refolding.AF2_PRIMARY_NATIVE_MOTIF_RMSD.key,
            descriptors_refolding.AF2_PRIMARY_PLDDT.key,
        ],
    )

    pool = get_cached_pool(design.pool_id)
    design_job = get_cached_design_job(pool.design_job_id)
    workflow: RFdiffusionScaffoldDesignWorkflow = design_job.workflow

    parser = ContigsParser()
    paths = (
        get_cached_design_descriptors(
            design_id,
            [
                descriptors_rfdiffusion.RFDIFFUSION_TRB_PATH.key,
                *[d.key for d in descriptors.STRUCTURE_PATH_DESCRIPTORS],
            ],
        )
        .dropna()
        .to_dict()
    )
    trb_dict = storage.read_file_pickle(paths[descriptors_rfdiffusion.RFDIFFUSION_TRB_PATH.key])
    input_pdb_str = storage.read_file_str(workflow.get_input_pdb_path(design.contig_index))
    input_segments = [segment for contig in trb_dict["sampled_mask"] for segment in parser.parse_contigs_str(contig)]
    output_segments = parser.parse_contigs_trb(trb_dict)

    input_mapping = [
        (segment.input_res_chain, list(range(segment.input_res_start, segment.input_res_end + 1)))
        for segment in input_segments
        if segment.type == "fixed"
    ]
    output_mapping = [
        (segment.out_res_chain, list(range(segment.out_res_start, segment.out_res_end + 1)))
        for segment in output_segments
        if segment.type == "fixed"
    ]

    left, middle, right = st.columns(3, gap="medium")

    with left:
        st.write("##### Full input structure")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=input_pdb_str, contigs=input_segments, representation_type="cartoon+ball-and-stick"
                ),
            ],
            key="full_input",
            height=350,
        )

    backbone_design_descriptor = descriptors_rfdiffusion.RFDIFFUSION_STRUCTURE_PATH

    with middle:
        st.write(f"##### {backbone_design_descriptor.name}")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=storage.read_file_str(paths[backbone_design_descriptor.key]),
                    contigs=output_segments,
                )
            ],
            key="rfdiff_contig_segments",
            height=350,
        )

        st.write(backbone_design_descriptor.description)

    sequence_design_descriptor = None
    for d in SEQUENCE_DESIGN_PATH_DESCRIPTORS:
        if paths.get(d.key):
            sequence_design_descriptor = d

    with right:
        st.write(f"##### {sequence_design_descriptor.name}")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=storage.read_file_str(paths[sequence_design_descriptor.key]),
                    contigs=output_segments,
                    representation_type="cartoon+ball-and-stick",
                )
            ],
            key="mpnn_contig_segments",
            height=350,
        )

        st.write(sequence_design_descriptor.description)

    st.write("### Refolding tests")

    prediction_descriptor = select_structure_prediction_descriptor(paths)

    if not prediction_descriptor:
        return

    left, middle, right = st.columns(3, gap="medium")

    prediction_storage_path = paths[prediction_descriptor.key]
    prediction_pdb = storage.read_file_str(prediction_storage_path)

    with left:
        st.write(f"##### {prediction_descriptor.name}")
        description = prediction_descriptor.description

        if prediction_descriptor.b_factor_value == "plddt":
            pdb_str = pdb_to_mmcif(prediction_pdb, "-", True)
            description += " (colored by pLDDT confidence)"
        elif prediction_descriptor.b_factor_value == "fractional_plddt":
            pdb_str = pdb_to_mmcif(prediction_pdb, "-", True, fractional_plddt=True)
            description += " (colored by pLDDT confidence)"
        else:
            pdb_str = prediction_pdb

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_str,
                    representation_type="cartoon",
                    color="plddt",
                ),
            ],
            key=f"pred_{prediction_descriptor.key}",
            height=350,
        )
        st.write(description)

    with middle:
        st.write(f"##### Input motif aligned to prediction")

        input_motif_pdb = filter_pdb_str(input_pdb_str, [s.value for s in input_segments if s.type == "fixed"])
        structures = [
            StructureVisualization(
                pdb=input_motif_pdb, contigs=input_segments, representation_type="cartoon+ball-and-stick"
            )
        ]

        (_, aligned_prediction_pdb), rmsd = align_multiple_proteins_pdb(
            pdb_strs=[input_motif_pdb, prediction_pdb],
            chain_residue_mappings=[
                input_mapping,
                output_mapping,
            ],
            all_atom=True,
        )
        num_fixed = sum(
            segment.out_res_end - segment.out_res_start + 1 for segment in output_segments if segment.type == "fixed"
        )

        structures.append(
            StructureVisualization(
                pdb=aligned_prediction_pdb,
                representation_type="cartoon+ball-and-stick",
            )
        )

        molstar_custom_component(
            structures=structures,
            key=f"input_motif_aligned_to_{prediction_descriptor.name}",
            height=350,
            html_filename=os.path.basename(prediction_storage_path).replace(".pdb", "_aligned_motif"),
        )

        st.write(
            f"""
            {prediction_descriptor.name} vs native motif all atom RMSD: **{rmsd:.2f} Å**

            {prediction_descriptor.name} prediction (white) aligned to original input structure (colored segments) 
            of the {num_fixed} fixed input motif residues.
            """
        )

    with right:
        st.write(f"##### Design aligned to prediction")

        structures, rmsd = align_multiple_proteins_pdb(
            pdb_strs=[
                storage.read_file_str(paths[sequence_design_descriptor.key]),
                prediction_pdb,
            ],
            chain_residue_mappings=[[("A", None)], [("A", None)]],
            all_atom=False,
        )
        aligned_str = structures[1]

        if prediction_descriptor.b_factor_value == "plddt":
            aligned_str = pdb_to_mmcif(aligned_str, "-", True)
        elif prediction_descriptor.b_factor_value == "fractional_plddt":
            aligned_str = pdb_to_mmcif(aligned_str, "-", True, fractional_plddt=True)

        molstar_custom_component(
            structures=[
                StructureVisualization(pdb=structures[0], contigs=output_segments),
                StructureVisualization(
                    pdb=aligned_str,
                    representation_type="cartoon",
                    color="plddt",
                ),
            ],
            key=f"design_aligned_to_{prediction_descriptor.key}",
            height=350,
        )

        st.write(
            f"""
            {prediction_descriptor.name} vs design backbone RMSD: **{rmsd:.2f} Å**
            
            Agreement between {sequence_design_descriptor.name} and {prediction_descriptor.name}
            """
        )


def rfdiffusion_multiple_binder_designs_visualization(all_design_ids: list[str], max_examples: int = 20):
    if len(all_design_ids) > max_examples:
        st.warning(
            f"Showing only first {max_examples} out of {len(all_design_ids):,} selected designs for performance reasons."
        )

    example_design_ids = all_design_ids[:max_examples]
    example_designs: list[Design] = get_cached_designs(example_design_ids)
    # TODO add caching
    backbone_paths = (
        db.select_descriptor_values(descriptors_rfdiffusion.RFDIFFUSION_STRUCTURE_PATH.key, example_design_ids)
        .dropna()
        .to_dict()
    )
    left, right = st.columns(2, gap="medium")

    with left:
        st.write("##### RFdiffusion backbone designs")
        aligned_pdb_strs, rmsd = align_multiple_proteins_pdb(
            pdb_strs=[storage.read_file_str(backbone_paths[design.id]) for design in example_designs],
            chain_residue_mappings=[[("B", None)] for _ in example_designs],
        )

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=aligned_pdb_str,
                    representation_type=None,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            color_params={"value": get_color_from_str(design.id, "seaborn:tab10_light")},
                            representation_type="cartoon",
                            label=f"RFdiffusion {design.id} binder backbone",
                        ),
                        ChainVisualization(
                            chain_id="B", representation_type="cartoon", label=f"RFdiffusion {design.id} target"
                        ),
                    ],
                )
                for design, aligned_pdb_str in zip(example_designs, aligned_pdb_strs)
            ],
            key="rfdiff_backbone",
        )


def rfdiffusion_binder_design_visualization(design_id: str):
    design: Design = get_cached_design(design_id)
    pool = get_cached_pool(design.pool_id)
    design_job = get_cached_design_job(pool.design_job_id)
    workflow: RFdiffusionBinderDesignWorkflow = design_job.workflow
    paths = (
        get_cached_design_descriptors(
            design_id,
            [
                descriptors_rfdiffusion.RFDIFFUSION_TRB_PATH.key,
                *[d.key for d in descriptors.STRUCTURE_PATH_DESCRIPTORS],
            ],
        )
        .dropna()
        .to_dict()
    )
    input_pdb_path = workflow.get_input_pdb_path(design.contig_index)

    st.write(f"Input structure: **{os.path.basename(input_pdb_path)}**")

    show_design_metrics(
        design_id,
        descriptor_keys=[
            descriptors_proteinqc.LENGTH.key,
            descriptors_rfdiffusion.RADIUS_OF_GYRATION.key,
            descriptors_rfdiffusion.PYDSSP_HELIX_PERCENT.key,
            descriptors_rfdiffusion.PYDSSP_SHEET_PERCENT.key,
            descriptors_rfdiffusion.N_CONTACTS_TO_HOTSPOTS.key,
            descriptors_rfdiffusion.PYROSETTA_DDG.key,
            descriptors_rfdiffusion.PYROSETTA_CMS.key,
            descriptors_rfdiffusion.PYROSETTA_SAP_SCORE.key,
            descriptors_refolding.AF2_PRIMARY_IPAE.key,
            descriptors_refolding.AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD.key,
            descriptors_refolding.AF2_PRIMARY_PLDDT.key,
        ],
    )

    # TODO use target spec for this
    # Note this assumes that the target chain is B in the RFdiffusion output PDB,
    # and that the residue numbers are same as in the input PDB
    hotspot_selections = (
        [h.replace(workflow.target_chain, "B") for h in workflow.rfdiffusion_params.hotspots.split(",")]
        if workflow.rfdiffusion_params.hotspots and workflow.target_chain
        else None
    )

    left, middle, right = st.columns(3, gap="medium")

    backbone_design_descriptor = descriptors_rfdiffusion.RFDIFFUSION_STRUCTURE_PATH
    with left:
        st.write(f"##### {backbone_design_descriptor.name}")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=storage.read_file_str(paths[backbone_design_descriptor.key]),
                    highlighted_selections=hotspot_selections,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            color_params={"value": get_color_from_str(design.id, "seaborn:tab10_light")},
                            representation_type="cartoon",
                            label=f"{backbone_design_descriptor.name} {design.id}",
                        ),
                        ChainVisualization(chain_id="B", representation_type="cartoon", label=f"{design.id} target"),
                    ],
                )
            ],
            key="backbone_design",
            height=350,
        )

        st.write(backbone_design_descriptor.description)

    sequence_design_descriptor = None
    for d in SEQUENCE_DESIGN_PATH_DESCRIPTORS:
        if paths.get(d.key):
            sequence_design_descriptor = d

    with middle:
        st.write(f"##### {sequence_design_descriptor.name}")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=storage.read_file_str(paths[sequence_design_descriptor.key]),
                    highlighted_selections=hotspot_selections,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            color_params={"value": get_color_from_str(design.id, "seaborn:tab10_light")},
                            representation_type="cartoon+ball-and-stick",
                            label=f"{sequence_design_descriptor.name} {design.id}",
                        ),
                        ChainVisualization(
                            chain_id="B", representation_type="cartoon+ball-and-stick", label=f"{design.id} target"
                        ),
                    ],
                )
            ],
            key="sequence_design_1",
            height=350,
        )

        st.write(sequence_design_descriptor.description)

    with right:
        st.write("##### Designed binding pose")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=storage.read_file_str(paths[sequence_design_descriptor.key]),
                    highlighted_selections=hotspot_selections,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            color_params={"value": get_color_from_str(design.id, "seaborn:tab10_light")},
                            representation_type="cartoon+ball-and-stick",
                            label=f"{sequence_design_descriptor.name} {design.id}",
                        ),
                        ChainVisualization(
                            chain_id="B",
                            representation_type="molecular-surface",
                            color="hydrophobicity",
                            label=f"{design.id} target",
                        ),
                    ],
                )
            ],
            key="sequence_design_2",
            height=350,
        )

        st.write(
            f"{sequence_design_descriptor.name} against target surface colored by hydrophobicity scale :green-badge[**green** = hydroPHOBIC] :red-badge[**red** = hydroPHILIC]"
        )

    st.write("### Refolding tests")

    prediction_descriptor = select_structure_prediction_descriptor(paths)

    if not prediction_descriptor:
        return

    left, middle, right = st.columns(3, gap="medium")
    with left:
        st.write("##### Input structure aligned to prediction")

        input_pdb_str = storage.read_file_str(input_pdb_path)
        prediction_str = storage.read_file_str(paths[prediction_descriptor.key])

        # here, manual alignment is needed
        structures, rmsd = align_multiple_proteins_pdb(
            pdb_strs=[input_pdb_str, prediction_str], chain_residue_mappings=[None, [("B", None)]]
        )
        aligned_str = structures[1]
        if prediction_descriptor.b_factor_value == "plddt":
            aligned_str = pdb_to_mmcif(aligned_str, "-", True)
        elif prediction_descriptor.b_factor_value == "fractional_plddt":
            aligned_str = pdb_to_mmcif(aligned_str, "-", True, fractional_plddt=True)

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=structures[0],
                    representation_type="cartoon",
                    color="chain-id",
                    highlighted_selections=workflow.selected_segments,
                ),
                StructureVisualization(
                    pdb=aligned_str,
                    representation_type=None,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            representation_type="cartoon+ball-and-stick",
                            color="plddt",
                            label=prediction_descriptor.name,
                        ),
                        ChainVisualization(
                            chain_id="B",
                            representation_type="cartoon",
                            label="Prediction of target chain aligned to input target chain",
                        ),
                    ],
                ),
            ],
            key="full_input_aligned_to_prediction",
            height=350,
        )

        st.write(
            f"All chains from input structure aligned to {prediction_descriptor.name} of binder chain (ball and stick colored by pLDDT confidence) and target chain (white)."
        )

    with middle:
        st.write("##### Design aligned to prediction")

        # here, we do not need manual alignment, but we still do it
        structures, rmsd = align_multiple_proteins_pdb(
            pdb_strs=[
                storage.read_file_str(paths[sequence_design_descriptor.key]),
                storage.read_file_str(paths[prediction_descriptor.key]),
            ],
            chain_residue_mappings=[[("B", None)] for _ in range(2)],
        )

        aligned_str = structures[1]
        if prediction_descriptor.b_factor_value == "plddt":
            aligned_str = pdb_to_mmcif(aligned_str, "-", True)

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=structures[0],
                    representation_type=None,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            color_params={"value": get_color_from_str(design.id, "seaborn:tab10_light")},
                            representation_type="cartoon+ball-and-stick",
                            label=sequence_design_descriptor.name,
                        ),
                        ChainVisualization(chain_id="B", representation_type="cartoon", label="MPNN target"),
                    ],
                ),
                StructureVisualization(
                    pdb=aligned_str,
                    representation_type=None,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            representation_type="cartoon+ball-and-stick",
                            color="plddt",
                            label=prediction_descriptor.name,
                        ),
                        ChainVisualization(chain_id="B", representation_type="cartoon", label="Target chain"),
                    ],
                ),
            ],
            key="design_aligned_to_af2_2",
            height=350,
        )

        st.write(
            f"""
            {prediction_descriptor.name} vs design backbone RMSD: **{rmsd:.2f} Å**
            
            Agreement between {sequence_design_descriptor.name} and {prediction_descriptor.name}
            """
        )

    with right:
        st.write("##### Predicted binding pose")

        pdb_str = storage.read_file_str(paths[prediction_descriptor.key])
        if prediction_descriptor.b_factor_value == "plddt":
            pdb_str = pdb_to_mmcif(pdb_str, "-", True)

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_str,
                    representation_type=None,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            representation_type="ball-and-stick",
                            color="plddt",
                            label="Prediction of binder chain",
                        ),
                        ChainVisualization(
                            chain_id="B",
                            color="hydrophobicity",
                            representation_type="molecular-surface",
                            label="Prediction of target chain",
                        ),
                    ],
                ),
            ],
            key="binding_pose_predicted",
            height=350,
        )

        st.write(
            "Prediction of binder (ball and stick colored by pLDDT confidence) and target (surface colored by hydrophobicity scale) :green-badge[**green** = hydroPHOBIC] :red-badge[**red** = hydroPHILIC]"
        )


def bindcraft_binder_design_visualization(design_id: str):
    design = get_cached_design(design_id)

    show_design_metrics(
        design_id,
        descriptor_keys=[
            "bindcraft|af2|Average_pLDDT",
            "bindcraft|af2|Average_i_pAE",
            "bindcraft|af2|Average_Hotspot_RMSD",
            "bindcraft|interface|Average_dG",
            "bindcraft|interface|Average_Relaxed_Clashes",
            "bindcraft|interface|Average_n_InterfaceResidues",
            "bindcraft|dssp|Average_Binder_Helix%",
            "bindcraft|dssp|Average_Binder_BetaSheet%",
        ],
    )

    molstar_custom_component(
        structures=[StructureVisualization(pdb=storage.read_file_str(design.structure_path))], key="bindcraft_1"
    )


def visualize_design_structure(design_id: str, height="500px"):
    design = get_cached_design(design_id)

    molstar_custom_component(
        structures=[
            StructureVisualization(
                pdb=storage.read_file_str(design.structure_path),
                color="chain-id",
                representation_type="cartoon+ball-and-stick",
            )
        ],
        key="default_structure",
        height=height,
    )


def visualize_design_sequence(design_id: str):
    design = get_cached_design(design_id)

    sequence_fasta = "\n".join(
        f">{design_id} {chain.type}|{','.join(chain.chain_ids)}\n{chain.sequence}" for chain in design.spec.chains
    )
    st.code(sequence_fasta, wrap_lines=True)


def visualize_rfdiffusion_design_sequence(design_id: str):
    # visualize regular sequence
    visualize_design_sequence(design_id)

    # visualize alignment
    st.write("##### Alignment of input sequence to designed sequence")

    design: Design = get_cached_design(design_id)
    pool = get_cached_pool(design.pool_id)
    workflow: RFdiffusionWorkflow = get_cached_design_job(pool.design_job_id).workflow

    input_pdb_path = workflow.get_input_pdb_path(design.contig_index)
    input_seq_by_resno: dict[str, dict[str, str]] = get_sequences_from_pdb_str(
        storage.read_file_str(input_pdb_path), by_residue_number=True
    )
    designed_sequences: dict[str, str] = {
        chain_id: chain.sequence for chain in design.spec.chains for chain_id in chain.chain_ids
    }
    paths = (
        get_cached_design_descriptors(design_id, [descriptors_rfdiffusion.RFDIFFUSION_TRB_PATH.key]).dropna().to_dict()
    )

    parser = ContigsParser()
    # TODO this could be parsed from spec contig instead
    parsed_segments = parser.parse_contigs_trb(
        storage.read_file_pickle(paths[descriptors_rfdiffusion.RFDIFFUSION_TRB_PATH.key])
    )
    # individual positions in this format ['A1', 'A2', ...]
    inpainted_positions = (
        from_segments_to_hotspots(workflow.rfdiffusion_params.inpaint_seq.split("/"))
        if workflow.rfdiffusion_params.inpaint_seq
        else []
    )
    visualize_scaffold_alignment(
        input_seq_by_resno=input_seq_by_resno,
        designed_sequences=designed_sequences,
        parsed_segments=parsed_segments,
        inpainted_positions=inpainted_positions,
    )


def visualize_scaffold_alignment(
    input_seq_by_resno: dict[str, dict[str, str]],
    designed_sequences: dict[str, str],
    parsed_segments: list[ContigSegment],
    inpainted_positions: list[str],
):
    """Visualize alignment between input and designed sequences based on contig segments.

    :param input_seq_by_resno: Mapping from chain ID -> residue number (as string) -> amino acid.
    :param designed_sequences: Mapping from chain ID -> designed sequence.
    :param parsed_segments: List of ContigSegment objects representing the contig segments.
    :param inpainted_positions: List of positions (e.g., 'A12') that were inpainted.
    """
    html = [
        f'<div style="font-family: monospace; font-size: 14px; margin: 10px 2px; display: inline-block;">'
        f"&nbsp;&nbsp;&nbsp;&nbsp;Region<br />"
        f"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Input<br />"
        f"&nbsp;&nbsp;&nbsp;&nbsp;Design<br />"
        f"</div>"
    ]
    aligned_input = []
    aligned_design = []
    for s in parsed_segments:
        segment_length = s.out_res_end - s.out_res_start + 1
        for chunk_offset in range(0, segment_length, 10):
            chunk_length = min(segment_length - chunk_offset, 10)
            output_start = chunk_offset + s.out_res_start
            # TODO s.out_res_chain might not be correct when multiple chains are designed
            generated_seq = designed_sequences[s.out_res_chain][output_start - 1 : output_start + chunk_length - 1]
            if s.type == "generated":
                label = s.value if chunk_offset == 0 else "&nbsp;"
                input_seq = " " * chunk_length
                input_seq_formatted = "Generated&nbsp;" if chunk_offset == 0 else "&nbsp;" * 10
                generated_seq_formatted = generated_seq
                fmt = f"background-color: {s.color}" if chunk_offset == 0 else ""
            else:
                input_start = chunk_offset + s.input_res_start
                label = f"{s.input_res_chain}{input_start}-{input_start + chunk_length - 1}"
                input_positions = list(range(input_start, input_start + chunk_length))
                input_seq = "".join(
                    [input_seq_by_resno[s.input_res_chain].get(str(pos), "?") for pos in input_positions]
                )
                input_seq_formatted = "".join(
                    [
                        f"<b>{aa}</b>" if f"{s.input_res_chain}{pos}" in inpainted_positions else aa
                        for pos, aa in zip(input_positions, input_seq)
                    ]
                )
                fmt = f"color: {s.color}"
                generated_seq_formatted = "".join(
                    [f"<b>{aa}</b>" if aa != bb else aa for aa, bb in zip(generated_seq, input_seq)]
                )
            aligned_input.extend(list(input_seq))
            aligned_design.extend(list(generated_seq))
            html.append(
                f'<div style="font-family: monospace; font-size: 14px; margin: 10px 2px; display: inline-block;">'
                f'<span style="display: inline-block; border-radius: 2px; padding: 1px 2px; {fmt}">{label.ljust(10).replace(" ", "&nbsp;")}</span><br />'
                f'<span style="color: #bbbbbb;">{input_seq_formatted}</span><br />'
                f"{generated_seq_formatted}<br />"
                f"</div>"
            )

    st.html("\n".join(html))

    if inpainted_positions:
        st.write("Sequence-inpainted positions shown in **bold**.")

    contig_identity = (
        sum(aa == bb for aa, bb in zip(aligned_input, aligned_design)) / len(aligned_design) if aligned_design else None
    )
    st.write(
        "Identity: **{:.2%}** :grey[:material/info: Not considering potential matches within generated structure regions]".format(
            contig_identity
        )
    )



import streamlit as st
import os
from ovo import config

ASSETS_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "assets")

CSS = None


def read_css():
    """Cache CSS contents to speed up page load"""
    global CSS
    if CSS is None:
        with open(os.path.join(ASSETS_PATH, "styles.css"), "r") as f:
            CSS = f.read()
    return CSS


@st.dialog("Welcome to OVO!")
def show_login_dialog(token: str = None):
    st.write("This OVO server requires login")
    if token:
        # Token-based login flow
        if st.query_params.get("login_token") == token:
            # Auto-login via URL param
            st.session_state.login_token = token
            st.rerun()
        # Manual token entry
        with st.form(border=False, key="login"):
            st.write("Please enter the login token to continue")
            token_input = st.text_input("Login Token", width="stretch")
            if st.form_submit_button("Continue", width="stretch", type="primary"):
                if token_input == token:
                    st.session_state.login_token = token
                    st.success("Login successful!")
                    st.rerun()
                else:
                    st.error("Invalid token, please try again.")
                    st.stop()
    else:
        # Streamlit login flow
        if st.button("Continue", width="stretch"):
            st.login()


def initialize_page(page_title=None):
    if st.session_state.get("project"):
        page_title += f" - {st.session_state.project.name}"
    st.set_page_config(
        layout="wide",
        page_title=page_title if "OVO" in page_title else page_title + " - OVO",
        page_icon=os.path.join(ASSETS_PATH, "ovo-favicons32x32.png"),
    )

    st.markdown(f"<style>{read_css()}</style>", unsafe_allow_html=True)

    if config.auth.streamlit_auth:
        if not st.user.get("is_logged_in"):
            show_login_dialog()
            st.stop()

    if login_token := os.environ.get("OVO_LOGIN_TOKEN"):
        if st.session_state.get("login_token") != login_token:
            show_login_dialog(token=login_token)
            st.stop()

    from ovo.app.components.sidebar import project_sidebar_component

    if st.session_state.get("flash_success"):
        st.success(st.session_state.pop("flash_success"))

    with st.spinner("Setting things up for you..."):
        if "workflows" not in st.session_state:
            st.session_state.workflows = {}

        if "initial_workflows" not in st.session_state:
            st.session_state.initial_workflows = {}

        if "pool_inputs" not in st.session_state:
            st.session_state.pool_inputs = {}

        if "new_project_name" not in st.session_state:
            st.session_state.new_project_name = None

        if "project" not in st.session_state:
            st.session_state.project = None  # filled in project_sidebar() below

        if "selection_history" not in st.session_state:
            st.session_state["selection_history"] = {}

        # Select project
        project_sidebar_component()



import hashlib
import os
import random
import re
import string
import uuid
from typing import Collection, Any

import numpy as np


def format_duration(td):
    """Returns: 1 day 2 hours and 5 minutes"""
    days = td.days
    hours, remainder = divmod(td.seconds, 3600)
    minutes, _ = divmod(remainder, 60)

    parts = []
    if days:
        parts.append(f"{days} day{'s' if days > 1 else ''}")
    if hours:
        parts.append(f"{hours} hour{'s' if hours > 1 else ''}")
    if minutes:
        parts.append(f"{minutes} minute{'s' if minutes > 1 else ''}")

    return " and ".join(parts) if parts else "0 minutes"


def generate_id(previous_ids: Collection[str]) -> str:
    """Generate a unique ID using 3 or more random letters.

    Args:
        previous_ids: List of existing IDs to avoid.

    Returns:
        str: A unique identifier.
    """
    # Choose length based on order of magnitude of previous IDs count, at least 3 characters
    # 0 -> 3, 9-9999 -> 3, 10000-99999 -> 4, 100000-999999 -> 5, etc.
    length = max(3, int(np.log10(len(previous_ids)) if previous_ids else 0))

    new_id = None
    while new_id is None or new_id in previous_ids:
        new_id = "".join(random.choices(string.ascii_letters.lower(), k=length))
    return new_id


def get_hash_of_bytes(value: bytes) -> str:
    """Get SHA1 hash string of the given bytes, for example '2aae6c35c94fcfb415dbe95f408b9ce91ee846ed'"""
    return hashlib.sha1(value).hexdigest()


def get_hashed_path_for_bytes(value: bytes) -> str:
    """Get directory and subdirectory based on SHA1 hash, for example '2a/ae6c35c94fcfb415dbe95f408b9ce91ee846ed'
                                                                         ^ Note the slash here

    This is done similarly to nextflow workdir to avoid exceeding the directory limit of some filesystems.
    """
    hash_str = get_hash_of_bytes(value)
    return os.path.join(hash_str[:2], hash_str[2:])


def safe_filename(filename):
    # Allow only the specified characters
    return re.sub(r"[^a-zA-Z0-9_.-]+", "_", filename).strip(".")


def parse_args(argv: list[str]) -> dict:
    """Parse list of args into a dict

    For example: ['--foo', 'bar', '--bar', 'baz', '--flag']
    into dict {'foo': 'bar', 'bar': 'baz', 'flag': True}
    """
    result = {}
    i = 0
    while i < len(argv):
        arg = argv[i]
        if arg.startswith("--"):
            key = arg.removeprefix("--")
            # If next is missing or is another flag → treat as boolean
            # Make sure to parse --run_params "--foo bar" as {"run_params": "--foo bar"} instead of {"run_params": True, "--foo bar": True}
            if i + 1 >= len(argv) or (argv[i + 1].startswith("--") and " " not in argv[i + 1]):
                result[key] = True
            else:
                result[key] = argv[i + 1]
                i += 1
        elif arg.startswith("-"):
            raise ValueError(f"Use double dashes --my_param value, got: {arg}")
        else:
            raise ValueError(f"Unexpected positional argument: {arg}")
        i += 1
    return result


def get_alphanumeric_sort_key(value: str) -> tuple:
    """Get a sort key that sorts strings with numbers in an alphanumeric way.

    For example, "item2" will come before "item10".

    :param value: The string to generate a sort key for.
    :return: A tuple that can be used as a sort key.
    """
    parts = re.split(r"(\d+)", value)
    return tuple([int(part) if part.isdigit() else (part or "").lower() for part in parts])


def sorted_alphanumeric(values: Collection[str]) -> list[str]:
    """Sort a list of strings in an alphanumeric  way.

    For example, "item2" will come before "item10".

    :param values: The list of strings to sort.
    :return: The sorted list of strings.
    """
    return sorted(values, key=get_alphanumeric_sort_key)


def truncated_list(items: Collection[Any], max_items: int, sep: str = ", ") -> str:
    """Join a list truncated to a maximum number of items, adding ellipsis if necessary."""
    if len(items) <= max_items:
        return sep.join(str(item) for item in items)
    else:
        return sep.join(str(item) for item in list(items)[:max_items]) + sep + "..."



import os
from copy import copy
from dataclasses import dataclass, field
from typing import Callable

from ovo.core.database.models import DesignWorkflow, WorkflowParams, WorkflowTypes, DesignJob, Design, Base
from ovo.core.scheduler.base_scheduler import Scheduler
from ovo.core.utils.residue_selection import from_segments_to_hotspots


@dataclass
class BindCraftParams(WorkflowParams):
    # path to input pdb in S3 or other storage backend
    input_pdb_path: str = field(default=None, metadata=dict(show_to_user=False, storage_path=True))
    # comma-separated target chain ids (for example "A,B")
    target_chains: str = None
    # minimum and maximum number of residues in designed binder
    binder_length: str = "70,150"
    # hotspot residues, comma-separated like A123,A124
    hotspots: str | None = None
    # run pipeline until given number of accepted designs are generated
    number_of_final_designs: int = 2
    # time limit in hours
    time_limit_hours: int = 24
    # filter (Default, Peptide, Relaxed, Peptide_Relaxed, None)
    filter_type: str = "Default"
    # design protocol (Default, Beta-sheet, Peptide)
    design_protocol: str = "Default"
    # interface protocol (AlphaFold2, MPNN)
    interface_protocol: str = "AlphaFold2"
    # template protocol (Default, Masked)
    template_protocol: str = "Default"
    # prediction protocol (Default, HardTarget)
    prediction_protocol: str = "Default"
    # dictionary with advanced settings that override default values
    # for example {"omit_AAs": ""}
    custom_advanced_settings: dict = None
    # dictionary with filter settings that override default values
    # for example {"1_pLDDT": {"threshold": 0.8, "higher": true}}
    custom_filter_settings: dict = None
    # Number of parallel replicas to run (default: 1). Each replica will stop when it reaches time_limit_seconds or when it generates the desired number of accepted designs.
    num_replicas: int = 1

    @classmethod
    def from_dict(cls, data):
        data = copy(data)
        return cls(**data)

    def validate(self):
        super().validate()
        if not self.input_pdb_path:
            raise ValueError("No input pdb provided")
        if not self.target_chains:
            raise ValueError("No target chains provided")
        if not self.binder_length:
            raise ValueError("No binder length provided")
        if not self.number_of_final_designs:
            raise ValueError("No number of final designs provided")
        if not self.binder_length.replace(",", "").isdigit():
            raise ValueError(f"Binder length should be 'length' or 'min,max', got: '{self.binder_length}'")
        if self.hotspots:
            # check if hotspots are in the form of A123,A124
            for hotspot in self.hotspots.split(","):
                if not hotspot:
                    raise ValueError(f"Invalid hotspot: {self.hotspots}")
                if not hotspot[0].isalpha():
                    raise ValueError(
                        f"Hotspot residue should be prefixed with chain ID (for example A123,A124), got: {self.hotspots}"
                    )
                if not any(hotspot.startswith(chain) for chain in self.target_chains.split(",")):
                    raise ValueError(f"Hotspot {hotspot} does not match selected target chains {self.target_chains}")


@WorkflowTypes.register("BindCraft binder design")
@dataclass
class BindCraftBinderDesignWorkflow(DesignWorkflow):
    """Dataclass for BindCraft binder design workflow."""

    # input parameters for BindCraft
    bindcraft_params: BindCraftParams | None = field(
        default_factory=BindCraftParams, metadata=dict(tool_name="BindCraft")
    )

    # UI state variables
    # User-provided PDB code (or filename in case of uploaded file)
    input_name: str | None = None
    # selected binder hotspots, synchronized with bindcraft_params.hotspots
    selected_segments: list[str] | None = None
    # dict chains, key: chain, value: contig, e.g. {A: A1-100/102-110/112-130, B: B23-202}
    chains: dict | None = None

    def get_input_pdb_paths(self):
        return [self.get_input_pdb_path()]

    def get_input_pdb_path(self):
        return self.bindcraft_params.input_pdb_path

    def get_selected_segments(self):
        return self.selected_segments or []

    def set_selected_segments(self, segments: list[str]):
        self.selected_segments = segments
        self.bindcraft_params.hotspots = from_segments_to_hotspots(segments)

    def get_pipeline_name(self) -> str:
        return "ovo.bindcraft"

    def prepare_params(self, workdir: str) -> dict:
        from ovo.core.logic.design_logic_bindcraft import prepare_bindcraft_params

        return prepare_bindcraft_params(self, workdir=workdir)

    def process_results(self, job: DesignJob, callback: Callable = None) -> list[Base]:
        """Process results of a successful workflow - download files from workdir,
        create Design and DescriptorValue objects"""
        from ovo.core.logic.design_logic_bindcraft import process_workflow_results

        return process_workflow_results(job=job, callback=callback)

    @classmethod
    def visualize_single_design_structures(cls, design_id: str):
        """Visualize single design structures in Streamlit"""
        from ovo.app.components.workflow_visualization_components import bindcraft_binder_design_visualization

        bindcraft_binder_design_visualization(design_id)

    @classmethod
    def get_download_fields(cls):
        return {
            "BindCraft Input PDB": (BindCraftBinderDesignWorkflow, "bindcraft_params.input_pdb_path"),
            "BindCraft design": (Design, "structure_path"),
        }

    def get_settings_paths(self):
        ovo_resources_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "resources"))
        bindcraft_resources_path = os.path.join(ovo_resources_path, "bindcraft")

        # Logic to determine protocol tags
        if self.bindcraft_params.design_protocol == "Default":
            design_protocol_tag = "default_4stage_multimer"
        elif self.bindcraft_params.design_protocol == "Beta-sheet":
            design_protocol_tag = "betasheet_4stage_multimer"
        elif self.bindcraft_params.design_protocol == "Peptide":
            design_protocol_tag = "peptide_3stage_multimer"
        else:
            raise ValueError(f"Unsupported design protocol {self.bindcraft_params.design_protocol}")

        if self.bindcraft_params.interface_protocol == "AlphaFold2":
            interface_protocol_tag = ""
        elif self.bindcraft_params.interface_protocol == "MPNN":
            interface_protocol_tag = "_mpnn"
        else:
            raise ValueError(f"Unsupported interface protocol {self.bindcraft_params.interface_protocol}")

        if self.bindcraft_params.template_protocol == "Default":
            template_protocol_tag = ""
        elif self.bindcraft_params.template_protocol == "Masked":
            template_protocol_tag = "_flexible"
        else:
            raise ValueError(f"Unsupported template protocol {self.bindcraft_params.template_protocol}")

        if self.bindcraft_params.design_protocol == "Peptide":
            prediction_protocol_tag = ""
        else:
            if self.bindcraft_params.prediction_protocol == "Default":
                prediction_protocol_tag = ""
            elif self.bindcraft_params.prediction_protocol == "HardTarget":
                prediction_protocol_tag = "_hardtarget"
            else:
                raise ValueError(f"Unsupported prediction protocol {self.bindcraft_params.prediction_protocol}")

        # Final advanced settings path
        advanced_settings_filename = (
            design_protocol_tag + interface_protocol_tag + template_protocol_tag + prediction_protocol_tag + ".json"
        )
        advanced_settings_path = os.path.join(bindcraft_resources_path, "settings_advanced", advanced_settings_filename)

        if self.bindcraft_params.filter_type == "Default":
            filter_filename = "default_filters.json"
        elif self.bindcraft_params.filter_type == "Peptide":
            filter_filename = "peptide_filters.json"
        elif self.bindcraft_params.filter_type == "Relaxed":
            filter_filename = "relaxed_filters.json"
        elif self.bindcraft_params.filter_type == "Peptide_Relaxed":
            filter_filename = "peptide_relaxed_filters.json"
        elif self.bindcraft_params.filter_type == "None":
            filter_filename = "no_filters.json"
        else:
            raise ValueError(f"Unsupported filter type: {self.bindcraft_params.filter_type}")

        filter_settings_path = os.path.join(bindcraft_resources_path, "settings_filters", filter_filename)

        return advanced_settings_path, filter_settings_path



from collections import Counter
from typing import List, Dict

import numpy as np
import pandas as pd
import streamlit as st

from ovo import CategoricalResidueDescriptor, ResidueNumberDescriptor, NumericGlobalDescriptor
from ovo.app.components.custom_elements import wrapped_columns
from ovo.app.components.descriptor_dialog import detail_button
from ovo.app.utils.protein_qc_plots import (
    format_threshold_value,
    descriptor_histograms,
    FLAG_ICONS,
)
from ovo.core.database import descriptors_rfdiffusion
from ovo.core.database.descriptors_proteinqc import (
    PROTEINQC_SEQUENCE_DESCRIPTORS,
    PROTEINQC_STRUCTURE_DESCRIPTORS,
    ESMFOLD_DESCRIPTORS,
    AF2_PRIMARY_DESCRIPTORS,
)
from ovo.core.database.models import Descriptor
from ovo.core.logic.proteinqc_logic import get_flag_color
from ovo.core.utils.formatting import truncated_list


@st.fragment()
def descriptor_overview_tiles(
    descriptors_df: pd.DataFrame,
    descriptors_by_key: Dict[str, Descriptor],
    design_id: str = None,
    histogram_source: str = "all",
):
    st.subheader(
        "Sequence-based descriptors",
        help="Overview of top sequence-based descriptors of the designs. The top histograms show distribution of the descriptor in PDB and the bottom histograms show distribution of the designs colored by preset warning and error thresholds.",
    )
    descriptor_tiles(
        descriptors_df, PROTEINQC_SEQUENCE_DESCRIPTORS, design_id=design_id, histogram_source=histogram_source
    )

    st.subheader(
        "Structure-based descriptors",
        help="Overview of top structure-based descriptors of the designs. The top histograms show distribution of the descriptor in PDB and the bottom histograms show distribution of the designs colored by preset warning and error thresholds.",
    )
    descriptor_tiles(
        descriptors_df, PROTEINQC_STRUCTURE_DESCRIPTORS, design_id=design_id, histogram_source=histogram_source
    )

    # AF2 initial guess
    af_descriptors = [descriptor for descriptor in AF2_PRIMARY_DESCRIPTORS if descriptor.key in descriptors_by_key]
    if af_descriptors:
        st.subheader("Refolding test: AlphaFold2 initial guess")
        descriptor_tiles(descriptors_df, af_descriptors, design_id=design_id, histogram_source=histogram_source)

    # ESMFold for scaffold design
    esmfold_descriptors = [descriptor for descriptor in ESMFOLD_DESCRIPTORS if descriptor.key in descriptors_by_key]
    if esmfold_descriptors:
        st.subheader("Refolding test: ESMFold")
        descriptor_tiles(descriptors_df, esmfold_descriptors, design_id=design_id, histogram_source=histogram_source)

    backbone_metrics_descriptors = [
        descriptor for descriptor in descriptors_rfdiffusion.BACKBONE_METRICS if descriptor.key in descriptors_by_key
    ]
    # FIXME: ADJUST HOTSPOT INTERFACE METRICS
    if backbone_metrics_descriptors:
        st.subheader("Backbone metrics")
        descriptor_tiles(
            descriptors_df, backbone_metrics_descriptors, design_id=design_id, histogram_source=histogram_source
        )


def get_residue_presence_df(descriptor: Descriptor, nonnull_descriptor_values, total_designs: int):
    """
    Returns a DataFrame with each unique residue and the percentage of designs in which it is present.
    """
    # Filter out empty or non-string values
    nonnull_values = [v for v in nonnull_descriptor_values if isinstance(v, str) and v.strip()]
    all_residues = []

    # Collect all residues from all designs by splitting on commas
    for v in nonnull_values:
        all_residues.extend([r.strip() for r in v.split(",") if r.strip()])
    if not all_residues:
        return None

    # Count in how many designs each residue is present
    residue_in_designs = Counter()
    for v in nonnull_values:
        residues = {r.strip() for r in v.split(",") if r.strip()}
        for residue in residues:
            residue_in_designs[residue] += 1
    residue_df = pd.DataFrame(
        {
            descriptor.name: list(residue_in_designs.keys()),
            "% of designs": [100 * count / total_designs for count in residue_in_designs.values()],
        }
    )
    residue_df = residue_df.sort_values(by="% of designs", ascending=False)
    return residue_df


def show_flag_counts(descriptor: NumericGlobalDescriptor, values: pd.Series, design_id: str = None):
    """Show a list of design counts by group (missing, green, yellow, orange)."""
    warning_ids = {"missing": [], "green": [], "yellow": [], "orange": []}
    if design_id:
        # only show results for selected design
        values = {design_id: values.get(design_id)}
    for d_id, val in values.items():
        flag = get_flag_color(val, descriptor)
        if not flag:
            continue
        warning_ids[flag].append(d_id)
    for flag, ids in warning_ids.items():
        if not ids:
            continue
        if flag == "missing":
            label = "missing value" if len(ids) == 1 else "missing values"
            color = "grey"
        else:
            label = f"{flag} flag" if len(ids) == 1 else f"{flag} flags"
            color = flag
        icon = FLAG_ICONS.get(flag, "")
        st.markdown(f":{color}-background[{icon} **{len(ids):,}** {label}]", help=truncated_list(ids, 10))


def descriptor_tiles(
    descriptors_df: pd.DataFrame, descriptors: List[Descriptor], design_id: str = None, histogram_source: str = "all"
):
    columns = wrapped_columns(len(descriptors), wrap=3, gap="medium")
    for col, descriptor in zip(columns, descriptors):
        with col:
            with st.container(border=True):
                st.markdown(f"#### {descriptor.name}", help=descriptor.description)
                if (descriptor.tool, descriptor.name) not in descriptors_df.columns:
                    st.warning("Not available")
                    continue
                descriptor_values = descriptors_df[(descriptor.tool, descriptor.name)]
                if isinstance(descriptor, NumericGlobalDescriptor):
                    descriptor_values = pd.to_numeric(descriptor_values, errors="coerce")

                nonnull_descriptor_values = descriptor_values.dropna().tolist()

                if not nonnull_descriptor_values:
                    st.warning("Not available")
                    continue

                if len(descriptor_values) > len(nonnull_descriptor_values):
                    perc_missing = (1 - len(nonnull_descriptor_values) / len(descriptor_values)) * 100
                    st.markdown(
                        f":grey-background[:material/question_mark:] **{perc_missing:.2f}%** values missing",
                    )

                if isinstance(descriptor, CategoricalResidueDescriptor):
                    top_n = 10
                    st.write(f"Top {top_n} most common values")
                    counts_df = pd.Series(nonnull_descriptor_values).value_counts().head(top_n).reset_index()
                    counts_df.columns = [descriptor.name, "Count"]
                    st.dataframe(
                        counts_df,
                        column_config={
                            descriptor.name: st.column_config.Column(descriptor.name),
                            "Count": st.column_config.ProgressColumn(
                                "Count",
                                format="%d",
                                min_value=0,
                                max_value=max(1, int(counts_df["Count"].max())),
                            ),
                        },
                        hide_index=True,
                        width="stretch",
                        height=200,
                    )
                    with st.container(horizontal=True, horizontal_alignment="right"):
                        detail_button(descriptor, descriptor_values, design_id=design_id)

                elif isinstance(descriptor, NumericGlobalDescriptor):
                    if design_id:
                        st.metric(
                            descriptor.name, format_threshold_value(descriptor.name, descriptor_values.loc[design_id])
                        )
                    else:
                        avg_value = np.mean(nonnull_descriptor_values)
                        st.metric(f"Average {descriptor.name}", format_threshold_value(descriptor.name, avg_value))

                    fig = descriptor_histograms(
                        descriptor=descriptor,
                        histogram_source=histogram_source,
                        values=nonnull_descriptor_values,
                        precomputed_histogram_height=25,
                        pool_histogram_height=70,
                        value=descriptor_values.loc[design_id] if design_id else None,
                        value_label=design_id,
                    )

                    st.altair_chart(fig, width="stretch")

                    with st.container(horizontal=True, horizontal_alignment="distribute", vertical_alignment="bottom"):
                        with st.container():
                            st.empty()
                            show_flag_counts(descriptor, descriptor_values, design_id=design_id)
                        detail_button(descriptor, descriptor_values, design_id=design_id)

                # Support for ResidueNumberDescriptor
                elif isinstance(descriptor, ResidueNumberDescriptor):
                    residue_df = get_residue_presence_df(descriptor, nonnull_descriptor_values, descriptors_df.shape[0])
                    if residue_df is None or residue_df.empty:
                        st.warning("No residue data available.")
                    else:
                        st.dataframe(
                            residue_df,
                            column_config={
                                descriptor.name: st.column_config.Column(descriptor.name),
                                "% of designs": st.column_config.ProgressColumn(
                                    "% of designs",
                                    format="%.1f%%",
                                    min_value=0,
                                    max_value=100,
                                    help="Percentage of designs in which the residue is present",
                                ),
                            },
                            hide_index=True,
                            width="stretch",
                            height=200,
                        )
                    with st.container(horizontal=True, horizontal_alignment="right"):
                        detail_button(descriptor, descriptor_values, design_id=design_id)
                else:
                    st.info("Visualization for this type of descriptor coming soon")
                    with st.container(horizontal=True, horizontal_alignment="right"):
                        detail_button(descriptor, descriptor_values, design_id=design_id)



import os
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from typing import Callable

from ovo import (
    db,
    storage,
    config,
    local_scheduler,
    get_scheduler,
    Design,
)
from ovo.core.database.models_rfdiffusion import (
    RFdiffusionWorkflow,
    RFdiffusionBinderDesignWorkflow,
    RFdiffusionScaffoldDesignWorkflow,
)
from ovo.core.database import descriptors_rfdiffusion, descriptors_refolding

from ovo.core.database.models import Pool, Round, DesignJob, DesignSpec, DescriptorValue, StructureFileDescriptor, Base
from ovo.core.logic.descriptor_logic import save_descriptor_job_for_design_job, read_descriptor_file_values
from ovo.core.logic.design_logic import set_designs_accepted


def submit_rfdiffusion_preview(
    workflow: RFdiffusionWorkflow,
    timesteps: int,
    partial_diffusion: bool = False,
    pipeline_name="rfdiffusion-backbone",
    **submission_args,
) -> str | None:
    """Run the RFdiffusion workflow with reduced number of diffuser timesteps."""

    contig = workflow.get_contig()
    hotspots = workflow.get_hotspots()

    if not contig:
        raise ValueError("Contig has not been computed. Please check the workflow parameters.")

    input_path = storage.prepare_workflow_input(workflow.get_input_pdb_path(), workdir=local_scheduler.workdir)

    run_parameters = []

    if partial_diffusion:
        run_parameters.append(f"diffuser.partial_T={timesteps}")
    else:
        run_parameters.append(f"diffuser.T={timesteps}")

    params = {
        "contig": contig,
        "input_pdb": os.path.abspath(input_path),
        "num_designs": 1,
        "hotspot": hotspots,
        "run_parameters": " ".join(run_parameters),
    }

    if workflow.get_cyclic_offset():
        params["cyclic"] = True

    preview_job_id = local_scheduler.submit(
        pipeline_name=pipeline_name,
        params=params,
        submission_args=submission_args,
    )
    return preview_job_id


def process_workflow_results(job: DesignJob, callback: Callable = None) -> list[Base]:
    pool = db.get(Pool, design_job_id=job.id)
    project_round = db.get(Round, id=pool.round_id)
    scheduler = get_scheduler(job.scheduler_key)
    workflow: RFdiffusionWorkflow = job.workflow
    assert workflow.is_instance(RFdiffusionWorkflow), (
        "This function expects a RFdiffusionWorkflow instance, got: {}".format(type(workflow).__name__)
    )

    # this is where result files will be stored in our storage
    # make sure to remove trailing slash otherwise S3 will keep two slashes in the path
    destination_dir = os.path.join("project", project_round.project_id, "pools", pool.id, "designs").rstrip("/")

    source_output_path = scheduler.get_output_dir(job.job_id)

    batch_size = int(workflow.rfdiffusion_params.batch_size)

    alphafold_file_suffix = None
    if workflow.refolding_params.primary_test and workflow.refolding_params.primary_test.startswith("af2_"):
        alphafold_file_suffix = workflow.refolding_params.primary_test

    esmfold_file_suffix = None
    if workflow.refolding_params.primary_test == "esmfold":
        esmfold_file_suffix = workflow.refolding_params.primary_test

    num_contigs = len(workflow.rfdiffusion_params.contigs)
    num_backbone_designs = workflow.rfdiffusion_params.num_designs
    num_sequence_designs = workflow.protein_mpnn_params.num_sequences
    num_fastrelax_cycles = workflow.protein_mpnn_params.fastrelax_cycles

    designs = []
    design_id_mapping = {}
    descriptor_values = []
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(
                process_rfdiffusion_design,
                pool_id=pool.id,
                batch_size=batch_size,
                contig_idx=contig_idx,
                num_contigs=num_contigs,
                total_idx_backbone=total_idx_backbone,
                num_backbone_designs=num_backbone_designs,
                num_sequence_designs=num_sequence_designs,
                num_fastrelax_cycles=num_fastrelax_cycles,
                source_output_path=source_output_path,
                destination_dir=destination_dir,
                alphafold_file_suffix=alphafold_file_suffix,
                esmfold_file_suffix=esmfold_file_suffix,
                cyclic=workflow.rfdiffusion_params.cyclic_offset,
            )
            for contig_idx in range(num_contigs)
            for total_idx_backbone in range(num_backbone_designs)
        ]

        for i, future in enumerate(futures):
            new_designs, new_mapping, new_values = future.result()
            designs.extend(new_designs)
            design_id_mapping.update(new_mapping)
            descriptor_values.extend(new_values)
            if callback and new_designs:
                callback(
                    value=(i + 1) / len(futures),
                    text=f"Downloading design {new_designs[0].id}",
                )

    # Create descriptor job on the fly
    descriptor_job = save_descriptor_job_for_design_job(
        design_job=job, project_id=project_round.project_id, chains=["A"], design_ids=list(design_id_mapping.keys())
    )
    for value in descriptor_values:
        value.descriptor_job_id = descriptor_job.id

    # Generate descriptor values from the descriptor output files
    filenames = {
        "proteinqc|seq_composition": "seq_composition",
        "rfd_ee|backbone_metrics": "backbone_metrics",
        "pyrosetta_interface_metrics|pyrosetta": "pyrosetta_interface_metrics",
    }
    if workflow.refolding_params.primary_test:
        # Store refolding results under the same set of Descriptor objects to simplify downstream analysis
        if workflow.refolding_params.primary_test.startswith("af2_"):
            descriptor_key_prefix = "refolding|af2_primary"
        else:
            descriptor_key_prefix = f"refolding|{workflow.refolding_params.primary_test}"
        # tool_key -> filename
        filenames[descriptor_key_prefix] = workflow.refolding_params.primary_test

    descriptor_values.extend(
        read_descriptor_file_values(
            descriptor_job=descriptor_job, design_id_mapping=design_id_mapping, filenames=filenames
        )
    )

    # Update design.accepted fields based on descriptor values and thresholds
    set_designs_accepted(designs, descriptor_values, workflow.acceptance_thresholds)
    # Return designs and descriptors to be saved
    return designs + descriptor_values


def process_rfdiffusion_design(
    pool_id: str,
    batch_size: int,
    contig_idx: int,
    num_contigs: int,
    total_idx_backbone: int,
    num_backbone_designs: int,
    num_sequence_designs: int,
    num_fastrelax_cycles: int,
    source_output_path: str,
    destination_dir: str,
    alphafold_file_suffix: str | None,
    esmfold_file_suffix: str | None,
    cyclic: bool,
) -> tuple[list[Design], dict[str, tuple[str, str]]]:
    batch_number = (total_idx_backbone // batch_size) + 1
    batch_name = f"contig{contig_idx + 1}_batch{batch_number}"
    batch_idx_backbone = total_idx_backbone % batch_size
    backbone_filename = f"{batch_name}_{batch_idx_backbone}"

    # add contig suffix 01 in case of multiple contigs
    contig_suffix = "_" + str(contig_idx + 1).zfill(max(len(str(num_contigs)), 2)) if num_contigs > 1 else ""
    # backbone suffix 01, 001, 0001 based on total number of designs
    backbone_suffix = "_" + str(total_idx_backbone + 1).zfill(max(len(str(num_backbone_designs)), 2))
    backbone_id = f"ovo_{pool_id}{contig_suffix}{backbone_suffix}"
    rfdiffusion_backbone_pdb_path = storage.store_file_path(
        source_abs_path=f"{source_output_path}/{batch_name}/rfdiffusion_standardized_pdb/{backbone_filename}_standardized.pdb",
        storage_rel_path=f"{destination_dir}/rfdiffusion/{backbone_id}_backbone.pdb",
        overwrite=False,
    )
    rfdiffusion_backbone_trb_path = storage.store_file_path(
        source_abs_path=f"{source_output_path}/{batch_name}/rfdiffusion_trb/{backbone_filename}.trb",
        storage_rel_path=f"{destination_dir}/rfdiffusion/{backbone_id}_backbone.trb",
        overwrite=False,
    )
    backbone_design = Design(
        id=backbone_id,
        pool_id=pool_id,
        accepted=False,
        contig_index=contig_idx,
    )
    designs = []
    design_id_mapping = {}
    descriptor_values = []

    if num_fastrelax_cycles > 0:
        mpnn_pdb_template = "proteinmpnn_fastrelax/{backbone_filename}_standardized_dldesign_0_cycle{idx_sequence}"
        seq_id_template = "_cycle{idx_sequence}"
        sequence_design_descriptor = descriptors_rfdiffusion.FASTRELAX_STRUCTURE_PATH
        num_seqs_total = num_fastrelax_cycles
    else:
        mpnn_pdb_template = "ligandmpnn/standardized_pdb/{backbone_filename}_standardized_packed_{num_sequence}_1"
        seq_id_template = "_seq{num_sequence}"
        sequence_design_descriptor = descriptors_rfdiffusion.LIGANDMPNN_STRUCTURE_PATH
        num_seqs_total = num_sequence_designs

    for idx_sequence in range(num_seqs_total):
        mpnn_source_path = mpnn_pdb_template.format(
            backbone_filename=backbone_filename,
            idx_sequence=idx_sequence,
            num_sequence=idx_sequence + 1,
        )
        filename = os.path.basename(mpnn_source_path)
        # create Design object
        design = deepcopy(backbone_design)
        design.id = backbone_id + seq_id_template.format(
            idx_sequence=str(idx_sequence).zfill(len(str(num_seqs_total))),
            # 01, 001 based on total number of sequences
            num_sequence=str(idx_sequence + 1).zfill(len(str(num_seqs_total))),
        )

        mpnn_full_source_path = os.path.join(source_output_path, batch_name, mpnn_source_path + ".pdb")
        if not storage.file_exists(mpnn_full_source_path):
            # skip designs that were filtered out before MPNN step
            # TODO read backbone_metrics.csv to know why they were filtered out,
            #  and only skip those with passed_filters=False, instead of checking file existence.
            #  The csv would need to be read from the correct batch folder, and only once per batch, not per design
            continue
        sequence_design_pdb_path = storage.store_file_path(
            source_abs_path=mpnn_full_source_path,
            storage_rel_path=f"{destination_dir}/protein_mpnn/{design.id}.pdb",
            overwrite=False,
        )
        design.structure_path = sequence_design_pdb_path
        design.structure_descriptor_key = sequence_design_descriptor.key
        design.spec = DesignSpec.from_pdb_str(
            pdb_data=storage.read_file_str(design.structure_path), chains=["A"], cyclic=cyclic
        )
        shared_args = dict(
            design_id=design.id,
            descriptor_job_id=None,
            chains="A",
        )
        descriptor_values.extend(
            [
                DescriptorValue(
                    descriptor_key=descriptors_rfdiffusion.RFDIFFUSION_STRUCTURE_PATH.key,
                    value=rfdiffusion_backbone_pdb_path,
                    **shared_args,
                ),
                DescriptorValue(
                    descriptor_key=descriptors_rfdiffusion.RFDIFFUSION_TRB_PATH.key,
                    value=rfdiffusion_backbone_trb_path,
                    **shared_args,
                ),
                DescriptorValue(
                    descriptor_key=sequence_design_descriptor.key,
                    value=sequence_design_pdb_path,
                    **shared_args,
                ),
            ]
        )

        if alphafold_file_suffix:
            descriptor_values.append(
                DescriptorValue(
                    descriptor_key=descriptors_refolding.AF2_PRIMARY_STRUCTURE_PATH.key,
                    value=storage.store_file_path(
                        source_abs_path=os.path.join(
                            source_output_path,
                            batch_name,
                            alphafold_file_suffix,
                            f"{filename}_{alphafold_file_suffix}.pdb",
                        ),
                        storage_rel_path=f"{destination_dir}/alphafold_initial_guess/{design.id}_{alphafold_file_suffix}.pdb",
                        overwrite=False,
                    ),
                    **shared_args,
                )
            )

        if esmfold_file_suffix:
            descriptor_values.append(
                DescriptorValue(
                    descriptor_key=descriptors_refolding.ESMFOLD_STRUCTURE_PATH.key,
                    value=storage.store_file_path(
                        source_abs_path=os.path.join(
                            source_output_path, batch_name, esmfold_file_suffix, f"{filename}_{esmfold_file_suffix}.pdb"
                        ),
                        storage_rel_path=f"{destination_dir}/esmfold/{design.id}_{esmfold_file_suffix}.pdb",
                        overwrite=False,
                    ),
                    **shared_args,
                )
            )

        designs.append(design)
        design_id_mapping[design.id] = (filename, backbone_filename + "_standardized")

    return designs, design_id_mapping, descriptor_values


def prepare_rfdiffusion_workflow_params(workflow: RFdiffusionWorkflow, workdir: str) -> dict:
    # prepare pdb file or txt file with multiple pdb paths
    workflow_input_path = storage.prepare_workflow_inputs(workflow.rfdiffusion_params.input_pdb_paths, workdir=workdir)
    design_type = workflow.get_refolding_design_type()
    params = {
        "batch_size": workflow.rfdiffusion_params.batch_size,
        "rfdiffusion_input_pdb": workflow_input_path,
        "rfdiffusion_num_designs": workflow.rfdiffusion_params.num_designs,
        "rfdiffusion_contig": ",".join(workflow.rfdiffusion_params.contigs),
        "rfdiffusion_run_parameters": get_rfdiffusion_run_parameters(workflow),
        "refolding_tests": workflow.refolding_params.primary_test,
        "design_type": design_type,
        "mpnn_num_sequences": workflow.protein_mpnn_params.num_sequences,
    }

    if workflow.rfdiffusion_params.backbone_filters:
        params["backbone_filters"] = workflow.rfdiffusion_params.backbone_filters

    # Disable ddG calculations if PyRosetta license is not available
    if not config.props.pyrosetta_license:
        params["disable_pyrosetta_scoring"] = True

    if workflow.protein_mpnn_params.fastrelax_cycles:
        # Use FastRelax
        if not config.props.pyrosetta_license:
            raise ValueError("FastRelax requires a PyRosetta license which is disabled in this instance of OVO.")
        params["mpnn_fastrelax_cycles"] = workflow.protein_mpnn_params.fastrelax_cycles
        params["mpnn_run_parameters"] = (
            f'-omit_AAs "{workflow.protein_mpnn_params.omit_aa}" '
            + f"-temperature {workflow.protein_mpnn_params.sampling_temp} "
            + (f'-bias_AA "{workflow.protein_mpnn_params.bias_aa}"' if workflow.protein_mpnn_params.bias_aa else "")
            + f" {workflow.protein_mpnn_params.run_parameters}"
        ).strip()
    else:
        # Otherwise use LigandMPNN
        params["mpnn_run_parameters"] = (
            f'--omit_AA "{workflow.protein_mpnn_params.omit_aa}" '
            + f"--temperature {workflow.protein_mpnn_params.sampling_temp}"
            + (f'--bias_AA "{workflow.protein_mpnn_params.bias_aa}"' if workflow.protein_mpnn_params.bias_aa else "")
        )

    if workflow.rfdiffusion_params.cyclic_offset:
        # Note that this is not supported by the public end-to-end workflow
        params["cyclic"] = True

    if workflow.rfdiffusion_params.hotspots:
        hotspots = ",".join(workflow.rfdiffusion_params.hotspots.replace(",", " ").split())
        params["hotspot"] = hotspots

    if workflow.refolding_params.esmfold_fp16:
        params["esmfold_fp16"] = True

    return params


def get_rfdiffusion_run_parameters(workflow: RFdiffusionWorkflow) -> str:
    args = ""
    if workflow.rfdiffusion_params.partial_diffusion:
        args += f" diffuser.partial_T={workflow.rfdiffusion_params.timesteps} "
    else:
        args += f" diffuser.T={workflow.rfdiffusion_params.timesteps} "

    if workflow.rfdiffusion_params.contigmap_length:
        length_range = (
            workflow.rfdiffusion_params.contigmap_length
            if "-" in str(workflow.rfdiffusion_params.contigmap_length)
            else f"{workflow.rfdiffusion_params.contigmap_length}-{workflow.rfdiffusion_params.contigmap_length}"
        )
        args += f" contigmap.length={length_range} "

    if workflow.rfdiffusion_params.inpaint_seq:
        args += f" contigmap.inpaint_seq=[{workflow.rfdiffusion_params.inpaint_seq}] "

    if workflow.rfdiffusion_params.model_weights not in [None, "Base", "Complex_base"]:
        args += f" inference.ckpt_override_path=rfdiffusion_models/{workflow.rfdiffusion_params.model_weights}_ckpt.pt "

    args += f" {workflow.rfdiffusion_params.run_parameters} "

    return args
